(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory AbstractRewriting imports Main begin

section {* Abstract Rewrite Systems *}

text {* An abstract rewrite system (ARS) is a binary relation. *}
types 'a brel = "('a \<times> 'a)set"

subsection {* Definitions *}

definition joinability :: "'a brel \<Rightarrow> 'a brel" ("(_^\<down>)" [1000] 999) where "r^\<down> \<equiv>  r^* O (r\<inverse>)^*"
definition meetability :: "'a brel \<Rightarrow> 'a brel" ("(_^\<up>)" [1000] 999) where "r^\<up> \<equiv> (r\<inverse>)^* O r^*"
abbreviation symmetric_closure :: "'a brel \<Rightarrow> 'a brel" ("(_^\<leftrightarrow>)" [1000] 999) where "r^\<leftrightarrow> \<equiv> r \<union> r\<inverse>"
definition conversion :: "'a brel \<Rightarrow> 'a brel" ("(_^\<leftrightarrow>*)" [1000] 999) where "r^\<leftrightarrow>* \<equiv> (r^\<leftrightarrow>)^*"

definition NF :: "'a brel \<Rightarrow> 'a set" where "NF r \<equiv> {x. r``{x} = {}}"

lemma no_step: assumes "r``{x} = {}" shows "x \<in> NF r" using assms by (auto simp: NF_def)

lemma joinability_I: "(x,y) \<in> r^* \<Longrightarrow> (z,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^\<down>"
  by (auto simp: joinability_def rtrancl_converse)

lemma joinability_I_left: "(x,y) \<in> r^* \<Longrightarrow> (x,y) \<in> r^\<down>" by (auto simp: joinability_def)

lemma joinability_I_right: "(y,x) \<in> r^* \<Longrightarrow> (x,y) \<in> r^\<down>" by (rule joinability_I) auto

declare joinability_I_left[intro]
declare joinability_I_right[intro]
declare joinability_I[intro]

lemma joinability_D[elim]:
assumes "(x,z) \<in> r^\<down>" obtains y where "(x,y) \<in> r^*" and "(z,y) \<in> r^*"
using assms by (auto simp: joinability_def rtrancl_converse)

lemma meetability_I[intro]: "(y,x) \<in> r^* \<Longrightarrow> (y,z) \<in> r^* \<Longrightarrow> (x,z) \<in> r^\<up>"
  by (auto simp: meetability_def rtrancl_converse)

lemma meetability_D[dest]: "(x,z) \<in> r^\<up> \<Longrightarrow> \<exists>y. (y,x) \<in> r^* \<and> (y,z) \<in> r^*"
  by (auto simp: meetability_def rtrancl_converse)

lemma meetability_E[elim]:
  "(y,z) \<in> r^\<up> \<Longrightarrow> (\<And>x. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> P) \<Longrightarrow> P" by auto

lemma conversion_I: "(x,y) \<in> (r^\<leftrightarrow>)^* \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>*" by (simp add: conversion_def)

lemma conversion_I': "(x,y) \<in> r^* \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>*" unfolding conversion_def
proof (induct rule: rtrancl_induct)
  case base thus ?case by simp
next
  case (step y z)
  hence "(y,z) \<in> r^\<leftrightarrow>" by simp
  with `(x,y) \<in> (r^\<leftrightarrow>)^*` show ?case by (rule rtrancl.intros(2))
qed

(* Later declarations are tried first for 'proof' and 'rule', hence
the "main" introduction / elimination rules for constants should be
declared last. *)
declare conversion_I'[intro]
declare conversion_I[intro]

lemma conversion_E[elim]: "(x,y) \<in> r^\<leftrightarrow>* \<Longrightarrow> ((x,y) \<in> (r^\<leftrightarrow>)^* \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: conversion_def)

lemma conversion_trans: "trans(r^\<leftrightarrow>*)" unfolding trans_def
proof (intro allI impI)
  fix x y z assume "(x,y) \<in> r^\<leftrightarrow>*" and "(y,z) \<in> r^\<leftrightarrow>*"
  thus "(x,z) \<in> r^\<leftrightarrow>*" unfolding conversion_def
  proof induct
    case base thus ?case by simp
  next
    case (step y z')
    from `(y,z') \<in> r^\<leftrightarrow>` and `(z',z) \<in> (r^\<leftrightarrow>)^*` have "(y,z) \<in> (r^\<leftrightarrow>)^*"
      by (rule converse_rtrancl_into_rtrancl)
    with step show ?case by simp
  qed
qed

lemma conversion_sym: "sym(r^\<leftrightarrow>*)" unfolding sym_def
proof (intro allI impI)
  fix x y assume "(x,y) \<in> r^\<leftrightarrow>*" thus "(y,x) \<in> r^\<leftrightarrow>*" unfolding conversion_def
  proof (induct)
    case base thus ?case by simp
  next
    case (step y z)
    hence "(z,y) \<in> r^\<leftrightarrow>" by auto
    from `(z,y) \<in> r^\<leftrightarrow>` and `(y,x) \<in> (r^\<leftrightarrow>)^*` show ?case by (rule converse_rtrancl_into_rtrancl)
  qed
qed

lemma joinability_left_I:
  assumes "(x,y) \<in> r^*" and "(y,z) \<in> r^\<down>" shows "(x,z) \<in> r^\<down>"
proof -
  from `(y,z) \<in> r^\<down>` obtain y' where "(y,y') \<in> r^*" and "(y',z) \<in> (r\<inverse>)^*"
    unfolding joinability_def by best
  with `(x,y) \<in> r^*` have "(x,y') \<in> r^*" by simp
  with `(y',z) \<in> (r\<inverse>)^*` show ?thesis unfolding joinability_def by auto
qed

lemma joinability_right_I:
  assumes "(x,y) \<in> r^\<down>" and "(z,y) \<in> r^*" shows "(x,z) \<in> r^\<down>"
proof -
  from `(z,y) \<in> r^*` have "(y,z) \<in> (r\<inverse>)^*" unfolding rtrancl_converse by simp
  from `(x,y) \<in> r^\<down>` obtain x' where "(x,x') \<in> r^*" and "(x',y) \<in> (r\<inverse>)^*"
    unfolding joinability_def by best
  with `(y,z) \<in> (r\<inverse>)^*` have "(x',z) \<in> (r\<inverse>)^*" by simp
  with `(x,x') \<in> r^*` show ?thesis unfolding joinability_def by auto
qed

lemma NF_I[intro]: "(\<And>y. (x,y) \<notin> r) \<Longrightarrow> x \<in> NF r" by (auto intro: no_step)

lemma NF_E[elim]: "x \<in> NF r \<Longrightarrow> ((\<And>y. (x,y) \<notin> r) \<Longrightarrow> P) \<Longrightarrow> P" by (auto simp: NF_def)

lemma NF_no_step: "x \<in> NF r \<Longrightarrow> \<forall>y. (x,y) \<notin> r" by auto

lemma NF_mono: assumes subset: "R \<subseteq> S" shows "NF S \<subseteq> NF R" 
proof
  fix x
  assume xS: "x \<in> NF S"
  show "x \<in> NF R"
  proof
    fix y
    show "(x,y) \<notin> R" using NF_no_step[OF xS] subset by auto
  qed
qed

lemma NF_iff_no_step: "t \<in> NF R = (\<forall> s. (t,s) \<notin> R)" by auto

lemma NF_no_trancl_step: assumes "x \<in> NF r" shows "\<forall>y. (x,y) \<notin> r^+"
proof -
  from assms have notin_r: "\<forall>y. (x,y) \<notin> r" by auto
  show ?thesis
  proof (intro allI notI)
    fix y assume "(x,y) \<in> r^+" thus False by induct (auto simp:notin_r)
   qed
qed

definition normalizability :: "'a brel \<Rightarrow> 'a brel" ("(_^!)" [1000] 999)
where "r^! \<equiv> {(x,y). (x,y) \<in> r^* \<and> y \<in> NF r}"

lemma normalizability_I: "(x,y) \<in> r^* \<Longrightarrow> y \<in> NF r \<Longrightarrow> (x,y) \<in> r^!"
  by (simp add: normalizability_def)

lemma normalizability_I': "(x,y) \<in> r^* \<Longrightarrow> (y,z) \<in> r^! \<Longrightarrow> (x,z) \<in> r^!"
  by (auto simp add: normalizability_def)

lemma normalizability_E: "(x,y) \<in> r^! \<Longrightarrow> ((x,y) \<in> r^* \<Longrightarrow> y \<in> NF r \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: normalizability_def)

declare normalizability_I'[intro]
declare normalizability_I[intro]
declare normalizability_E[elim]


subsection {* Properties of ARSs *}

text {*
The following properties on (elements of) ARSs are defined: complete, Church-Rosser property,
semi-complete, strong normalization, unique normal forms, weak Church-Rosser property,
weak normalization. 
*}

definition CR_elt where "CR_elt r x \<equiv> (\<forall>y z. (x,y) \<in> r^* \<and> (x,z) \<in> r^* \<longrightarrow> (y,z) \<in> r^\<down>)"
definition CR where "CR r \<equiv> (\<forall>x. CR_elt r x)"
definition SN_elt where "SN_elt r x \<equiv> \<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r))"
definition SN where "SN r \<equiv> \<forall>x. SN_elt r x"
definition UNF_elt where "UNF_elt r x \<equiv> \<forall>y z. (x,y) \<in> r^! \<and> (x,z) \<in> r^! \<longrightarrow> y = z"
definition UNF where "UNF r \<equiv> \<forall>x. UNF_elt r x"
definition WCR_elt where "WCR_elt r x \<equiv> \<forall>y z. (x,y) \<in> r \<and> (x,z) \<in> r \<longrightarrow> (y,z) \<in> r^\<down>"
definition WCR where "WCR r \<equiv> \<forall>x. WCR_elt r x"
definition WN_elt where "WN_elt r x \<equiv> \<exists>y. (x,y) \<in> r^!"
definition WN where "WN r \<equiv> \<forall>x. WN_elt r x"

lemmas CR_defs = CR_def CR_elt_def
lemmas SN_defs = SN_def SN_elt_def
lemmas UNF_defs = UNF_def UNF_elt_def
lemmas WCR_defs = WCR_def WCR_elt_def
lemmas WN_defs = WN_def WN_elt_def

lemma SN_def': "SN r \<equiv> \<not>(\<exists>S.\<forall>i. (S i,S(Suc i)) \<in> r)" by (simp add: SN_defs)

definition complete_elt where "complete_elt r x \<equiv> SN_elt r x \<and> CR_elt r x"
definition complete where "complete r \<equiv> \<forall>x. complete_elt r x"
definition semi_complete_elt where "semi_complete_elt r x \<equiv>  WN_elt r x \<and> CR_elt r x"
definition semi_complete where "semi_complete r \<equiv> \<forall>x. semi_complete_elt r x"

lemmas complete_defs = complete_def complete_elt_def
lemmas semi_complete_defs = semi_complete_def semi_complete_elt_def

text {* Unique normal forms with respect to conversion. *}
definition UNC :: "'a brel \<Rightarrow> bool"
where "UNC r \<equiv> \<forall>x y. x \<in> NF r \<and> y \<in> NF r \<and> (x,y) \<in> r^\<leftrightarrow>* \<longrightarrow> x = y"

lemma complete_elt_I[intro]: "SN_elt r x \<Longrightarrow> CR_elt r x \<Longrightarrow> complete_elt r x"
  by (simp add: complete_defs)

lemma complete_elt_E[elim]:
  "complete_elt r x \<Longrightarrow> (SN_elt r x \<Longrightarrow> CR_elt r x \<Longrightarrow> P) \<Longrightarrow> P" by (simp add: complete_defs)

lemma complete_I: "(\<And>x. SN_elt r x) \<Longrightarrow> (\<And>x. CR_elt r x) \<Longrightarrow> complete r"
  by (simp add: complete_defs)

lemma complete_I': "(\<And>x. complete_elt r x) \<Longrightarrow> complete r"
  by (simp add: complete_defs)

declare complete_I'[intro]
declare complete_I[intro]

lemma complete_E: "complete r \<Longrightarrow> (SN r \<Longrightarrow> CR r \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: complete_defs SN_defs CR_defs)

lemma complete_E': "complete r \<Longrightarrow> ((\<And>x. complete_elt r x) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: complete_defs)

lemma complete_E'':
  "complete r \<Longrightarrow> ((\<And>x. SN_elt r x \<Longrightarrow> CR_elt r x) \<Longrightarrow> P) \<Longrightarrow> P" by (simp add: complete_defs)

declare complete_E''[elim]
declare complete_E'[elim]
declare complete_E[elim]

lemma CR_elt_I[intro]:"(\<And>y z. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> CR_elt r x"
  by (simp add: CR_defs)

lemma CR_elt_E[elim]:
  "CR_elt r x \<Longrightarrow> ((\<And>y z. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: CR_defs)

lemma CR_I: "(\<And>x. CR_elt r x) \<Longrightarrow> CR r" by (simp add: CR_defs)

lemma CR_I': "(\<And>x y z. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> CR r"
  by (auto simp: CR_defs)

lemma CR_E: "CR r \<Longrightarrow> ((\<And>x. CR_elt r x) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: CR_defs)

lemma CR_E':
  "CR r \<Longrightarrow> ((\<And>x y z. (x,y) \<in> r^* \<Longrightarrow> (x,z) \<in> r^* \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> P) \<Longrightarrow> P"
  by (auto simp: CR_defs)

declare CR_I'[intro]
declare CR_I[intro]
declare CR_E'[elim]
declare CR_E[elim]

lemma semi_complete_elt_I[intro]: "WN_elt r x \<Longrightarrow> CR_elt r x \<Longrightarrow> semi_complete_elt r x"
  by (simp add: semi_complete_defs)

lemma semi_complete_elt_E[elim]:
  "semi_complete_elt r x \<Longrightarrow> (WN_elt r x \<Longrightarrow> CR_elt r x \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: semi_complete_defs)

lemma semi_complete_I: "(\<And>x. semi_complete_elt r x) \<Longrightarrow> semi_complete r"
  by (simp add: semi_complete_defs)

lemma semi_complete_I': "(\<And>x. WN_elt r x \<and> CR_elt r x) \<Longrightarrow> semi_complete r"
  by (simp add: semi_complete_defs)

lemma semi_complete_E:
  "semi_complete r \<Longrightarrow> ((\<And>x. semi_complete_elt r x) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: semi_complete_defs)

lemma semi_complete_E':
  "semi_complete r \<Longrightarrow> ((\<And>x. WN_elt r x) \<Longrightarrow> (\<And>x. CR_elt r x) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: semi_complete_defs)

lemma semi_complete_E'':
  "semi_complete r \<Longrightarrow> (WN r \<Longrightarrow> CR r \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: semi_complete_defs WN_defs CR_defs)

declare semi_complete_I'[intro]
declare semi_complete_I[intro]
declare semi_complete_E''[elim]
declare semi_complete_E'[elim]
declare semi_complete_E[elim]

lemma UNC_I[intro]: "(\<And>x y. x \<in> NF r \<Longrightarrow> y \<in> NF r \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>* \<Longrightarrow> x = y) \<Longrightarrow> UNC r"
  by (simp add: UNC_def)

lemma UNC_E[elim]:
  "UNC r \<Longrightarrow> ((\<And>x y. x \<in> NF r \<Longrightarrow> y \<in> NF r \<Longrightarrow> (x,y) \<in> r^\<leftrightarrow>* \<Longrightarrow> x = y) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: UNC_def)

lemma UNF_elt_I[intro]: "(\<And>y z. (x,y) \<in> r^! \<Longrightarrow> (x,z) \<in> r^! \<Longrightarrow> y = z) \<Longrightarrow> UNF_elt r x"
  by (simp add: UNF_defs)

lemma UNF_elt_E[elim]:
 "UNF_elt r x \<Longrightarrow> ((\<And>y z. (x,y) \<in> r^! \<Longrightarrow> (x,z) \<in> r^! \<Longrightarrow> y = z) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: UNF_defs)


lemma UNF_I: "(\<And>x y z. (x,y) \<in> r^! \<Longrightarrow> (x,z) \<in> r^! \<Longrightarrow> y = z) \<Longrightarrow> UNF r"
  by (auto simp: UNF_defs)

lemma UNF_I': "(\<And>x. UNF_elt r x) \<Longrightarrow> UNF r" by (simp add: UNF_defs)

declare UNF_I'[intro]
declare UNF_I[intro]

lemma SN_elt_I[intro]: "\<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)) \<Longrightarrow> SN_elt r x"
  by (simp add: SN_defs)

lemma SN_elt_E[elim]:
  "SN_elt r x \<Longrightarrow> ((\<And>S. S 0 = x \<Longrightarrow> (\<exists>i. (S i,S(Suc i)) \<notin> r)) \<Longrightarrow> P) \<Longrightarrow> P" by (simp add: SN_defs)

lemma SN_I[intro]: "(\<And>x. SN_elt r x) \<Longrightarrow> SN r" by (simp add: SN_defs)

lemma SN_E:
  "SN r \<Longrightarrow> (\<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)) \<Longrightarrow> P) \<Longrightarrow> P" by (simp add: SN_defs)

lemma SN_E': "SN r \<Longrightarrow> ((\<And>x. SN_elt r x) \<Longrightarrow> P) \<Longrightarrow> P" by (simp add: SN_defs)

declare SN_E'[elim]
declare SN_E[elim]

lemma WCR_elt_I[intro]: "(\<And>y z. (x,y) \<in> r \<Longrightarrow> (x,z) \<in> r \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> WCR_elt r x"
  by (simp add: WCR_defs)

lemma WCR_elt_E[elim]:
  "WCR_elt r x \<Longrightarrow> ((\<And>y z. (x,y) \<in> r \<Longrightarrow> (x,z) \<in> r \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: WCR_defs)

lemma WCR_I: "(\<And>x. WCR_elt r x) \<Longrightarrow> WCR r" by (simp add: WCR_defs)

lemma WCR_I': "(\<And>x y z. (x,y) \<in> r \<Longrightarrow> (x,z) \<in> r \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> WCR r"
  by (auto simp: WCR_defs)

lemma WCR_E: "WCR r \<Longrightarrow> ((\<And>x. WCR_elt r x) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: WCR_defs)

lemma WCR_E':
  "WCR r \<Longrightarrow> ((\<And>x y z. (x,y) \<in> r \<Longrightarrow> (x,z) \<in> r \<Longrightarrow> (y,z) \<in> r^\<down>) \<Longrightarrow> P) \<Longrightarrow> P"
  by (auto simp: WCR_defs)

declare WCR_I'[intro]
declare WCR_I[intro]
declare WCR_E'[elim]
declare WCR_E[elim]

lemma WN_elt_I[intro]: "(x,y) \<in> r^! \<Longrightarrow> WN_elt r x" by (auto simp: WN_defs)

lemma WN_elt_E[elim]: "WN_elt r x \<Longrightarrow> (\<exists>y.(x,y) \<in> r^! \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: WN_defs)

lemma WN_I[intro]: "(\<And>x. WN_elt r x) \<Longrightarrow> WN r" by (simp add: WN_defs)

lemma WN_E[elim]: "WN r \<Longrightarrow> ((\<And>x. WN_elt r x) \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp add: WN_defs)

definition SN_rel :: "'a brel \<Rightarrow> 'a brel \<Rightarrow> 'a brel" 
where "SN_rel R S \<equiv> {(s,t) | s t. (s,t) \<in> R \<and> SN_elt S s}"

lemma SN_SN_rel: "SN(SN_rel R R)" by (auto simp: SN_rel_def SN_defs)

lemma step_preserves_SN:
  fixes R :: "'a brel"
  assumes "(s,t) \<in> R" and "SN_elt R s" shows "SN_elt R t"
proof (rule ccontr)
 assume "\<not> SN_elt R t"
 then obtain A where "A 0 = t" and "(\<forall>i. (A i,A(Suc i)) \<in> R)" unfolding SN_defs by best
 let ?B = "\<lambda>i. case i of 0 \<Rightarrow> s | Suc i \<Rightarrow> A i"
 have "\<forall>i. (?B i,?B(Suc i)) \<in> R"
 proof
  fix i
  show "(?B i,?B (Suc i)) \<in> R"
  proof (cases i)
   assume "i = 0" with `A 0 = t` and `(s,t) \<in> R` show ?thesis by simp
  next
   fix j assume "i = Suc j" 
   from `\<forall>i. (A i,A(Suc i)) \<in> R` have "(A j,A (Suc j)) \<in> R" by simp
   with `i = Suc j` show ?thesis by simp
  qed
 qed
 hence "?B 0 = s \<and> (\<forall>i. (?B i,?B (Suc i)) \<in> R)" by simp
 hence "\<exists>B. B 0 = s \<and> (\<forall>i. (B i,B(Suc i)) \<in> R)" by best
 hence "\<not> SN_elt R s" unfolding SN_defs by simp
 with assms show False by simp
qed

lemma rtrancl_step_preserves_SN: "(s,t) \<in> R^* \<Longrightarrow> SN_elt R s \<Longrightarrow> SN_elt R t"
  by (induct rule: rtrancl.induct) (auto simp: step_preserves_SN)

lemma subsetI2:"(\<And>x y. (x,y) \<in> A \<Longrightarrow> (x,y) \<in> B) \<Longrightarrow> A \<subseteq> B" by auto

lemma SN_rel_trancl_simp[simp]: "(SN_rel R R)^+ = SN_rel (R^+) R" (is "?R = ?S")
proof
  show "?R \<subseteq> ?S"
  proof (rule subsetI2)
    fix x y assume "(x,y) \<in> ?R" thus "(x,y) \<in> ?S"
      unfolding SN_rel_def by (induct rule: trancl.induct) auto
  qed
next
  show "?S \<subseteq> ?R"
  proof (rule subsetI2)
    fix x y assume "(x,y) \<in> ?S"
    hence "(x,y) \<in> R^+" and "SN_elt R x" unfolding SN_rel_def by auto
    thus "(x,y) \<in> ?R"
    proof (induct rule: trancl.induct)
      case (r_into_trancl x y) thus ?case unfolding SN_rel_def by auto
    next
      case (trancl_into_trancl x y z)
      hence IH: "(x,y) \<in> ?R" by auto
      from trancl_into_trancl have "(x,y) \<in> R^*" by auto
      from this and  `SN_elt R x` have "SN_elt R y" by (rule rtrancl_step_preserves_SN)
      with `(y,z) \<in> R` have "(y,z) \<in> ?R" unfolding SN_rel_def by auto
      with IH show ?case by simp
    qed
  qed
qed

(* Only used locally and hidden afterwards. *)
primrec S :: "'a brel \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a"
where "S r x 0 = x"
    | "S r x (Suc n) = (\<some> y. (S r x n, y) \<in> r)"

lemma SN_imp_WN: assumes "SN r" shows "WN r"
proof -
 have "\<forall>x. WN_elt r x"
 proof
  fix x
  from assms have "\<not>(\<exists> S. S 0 = x \<and> (\<forall>i. (S i,(S(Suc i))) \<in> r))" by blast
  hence "\<exists>y. (x,y) \<in> {(x,y). (x,y) \<in> r^* \<and> y \<in> NF r}"
  proof (rule contrapos_np)
   let ?S = "\<lambda>i. S r x i"
   assume "\<not>(\<exists>y. (x,y) \<in> {(x,y). (x,y) \<in> r^* \<and> y \<in> NF r})"
   hence all:"\<forall>y. (x,y) \<in> r^* \<longrightarrow> (\<exists>z. (y,z) \<in> r)" by auto
   have "?S 0 = x \<and> (\<forall>i. (x,?S(Suc i)) \<in> r^^Suc i \<and> (?S i,?S(Suc i)) \<in> r)"
   proof
    show "?S 0 = x" by simp
   next
    show "\<forall>i. (x,?S(Suc i)) \<in> r^^Suc i \<and> (?S i,?S(Suc i)) \<in> r"
    proof
     fix i show "(x,?S(Suc i)) \<in> r^^Suc i \<and> (?S i,?S(Suc i)) \<in> r"
     proof (induct i)
      case 0 thus ?case by (auto simp:all intro:someI_ex)
     next
      case (Suc n)
      hence IH: "(x,?S(Suc n)) \<in> r^^Suc n" and "(?S n,?S(Suc n)) \<in> r" by auto
      hence "(x,?S(Suc n)) \<in> r^*" using rel_pow_imp_rtrancl by best
      with all have "\<exists>y. (?S(Suc n),y) \<in> r" by auto
      hence "(?S(Suc n),?S(Suc(Suc n))) \<in> r" by (auto intro:someI_ex)
      with IH show ?case using rel_pow_imp_rtrancl by auto
     qed
    qed
   qed
   thus "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by best
  qed
  thus "WN_elt r x" by auto
 qed
 thus ?thesis by auto
qed

hide const S

lemma UNC_imp_UNF:
 assumes "UNC r" shows "UNF r"
proof - {
 fix x y z assume "(x,y) \<in> r^!" and "(x,z) \<in> r^!"
 hence "(x,y) \<in> r^*" and "(x,z) \<in> r^*" and "y \<in> NF r" and "z \<in> NF r" by auto
 hence "(x,y) \<in> r^\<leftrightarrow>*" and "(x,z) \<in> r^\<leftrightarrow>*" by (auto intro: conversion_I')
 hence "(z,x) \<in> r^\<leftrightarrow>*" using conversion_sym unfolding sym_def by best
 with `(x,y) \<in> r^\<leftrightarrow>*` have "(z,y) \<in> r^\<leftrightarrow>*" using conversion_trans unfolding trans_def by best
 with `z \<in> NF r` and `y \<in> NF r` have "y = z" using assms by auto
} thus ?thesis by auto
qed

lemma joinability_NF_imp_eq:
 assumes "(x,y) \<in> r^\<down>" and "x \<in> NF r" and "y \<in> NF r"
 shows "x = y"
proof -
 from `(x,y) \<in> r^\<down>` obtain z where "(x,z)\<in>r^*" and "(z,y)\<in>(r\<inverse>)^*" unfolding joinability_def by auto
 hence "(y,z) \<in> r^*" unfolding rtrancl_converse by simp
 from `x \<in> NF r` have "(x,z) \<notin> r^+" using NF_no_trancl_step by best
 hence "x = z" using rtranclD[OF `(x,z) \<in> r^*`] by auto
 from `y \<in> NF r` have "(y,z) \<notin> r^+" using NF_no_trancl_step by best
 hence "y = z" using rtranclD[OF `(y,z) \<in> r^*`] by auto
 with `x = z` show ?thesis by simp
qed

lemma CR_iff_meetability_subset_joinability: "CR r = (r^\<up> \<subseteq> r^\<down>)"
proof
 assume "CR r" show "r^\<up> \<subseteq> r^\<down>"
 proof (rule subsetI2)
  fix x y assume "(x,y) \<in> r^\<up>"
  then obtain z where "(z,x) \<in> r^*" and "(z,y) \<in> r^*" using meetability_D by best
  with `CR r` show "(x,y) \<in> r^\<down>" by (auto simp: CR_defs)
 qed
next
 assume "r^\<up> \<subseteq> r^\<down>" {
  fix x y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
  hence "(y,z) \<in> r^\<up>" unfolding meetability_def rtrancl_converse by auto
  with `r^\<up> \<subseteq> r^\<down>` have "(y,z) \<in> r^\<down>" by auto
 } thus "CR r" by (auto simp: CR_defs)
qed

lemma CR_divergence_imp_joinability:
  assumes "CR r" and "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
  shows "(y,z) \<in> r^\<down>"
  using assms by (auto elim: CR_E')

lemma joinability_imp_conversion: "r^\<down> \<subseteq> r^\<leftrightarrow>*"
proof (rule subsetI2)
  fix x z assume "(x,z) \<in> r^\<down>"
  then obtain y where "(x,y) \<in> r^*" and "(z,y) \<in> r^*" by auto
  hence "(x,y) \<in> r^\<leftrightarrow>*" and "(z,y) \<in> r^\<leftrightarrow>*" by (auto intro: conversion_I')
  from `(z,y) \<in> r^\<leftrightarrow>*` have "(y,z) \<in> r^\<leftrightarrow>*" using conversion_sym unfolding sym_def by best
  with `(x,y) \<in> r^\<leftrightarrow>*` show "(x,z) \<in> r^\<leftrightarrow>*" using conversion_trans unfolding trans_def by best
qed

lemma meetability_imp_conversion: "r^\<up> \<subseteq> r^\<leftrightarrow>*"
proof (rule subsetI2)
  fix y z assume "(y,z) \<in> r^\<up>"
  then obtain x where "(x,y) \<in> r^*" and "(x,z) \<in> r^*" by auto
  hence "(x,y) \<in> r^\<leftrightarrow>*" and "(x,z) \<in> r^\<leftrightarrow>*" by (auto intro: conversion_I')
  from `(x,y) \<in> r^\<leftrightarrow>*` have "(y,x) \<in> r^\<leftrightarrow>*" using conversion_sym unfolding sym_def by best
  with `(x,z) \<in> r^\<leftrightarrow>*` show "(y,z) \<in> r^\<leftrightarrow>*" using conversion_trans unfolding trans_def by best
qed

lemma CR_imp_UNF: assumes "CR r" shows "UNF r"
proof - {
fix x y z assume "(x,y) \<in> r^!" and "(x,z) \<in> r^!"
  hence "(x,y) \<in> r^*" and "y \<in> NF r" and "(x,z) \<in> r^*" and "z \<in> NF r"
    unfolding normalizability_def by auto
  from assms and `(x,y) \<in> r^*` and `(x,z) \<in> r^*` have "(y,z) \<in> r^\<down>"
    by (rule CR_divergence_imp_joinability)
  from this and `y \<in> NF r` and `z \<in> NF r` have "y = z" by (rule joinability_NF_imp_eq)
} thus ?thesis by auto
qed

lemma CR_iff_conversion_imp_joinability: "CR r = (r^\<leftrightarrow>* \<subseteq> r^\<down>)"
proof (intro iffI subsetI2)
  fix x y assume "CR r" and "(x,y) \<in> r^\<leftrightarrow>*"
  then obtain n where "(x,y) \<in> (r^\<leftrightarrow>)^^n" unfolding conversion_def rtrancl_is_UN_rel_pow by auto
  thus "(x,y) \<in> r^\<down>"
  proof (induct n arbitrary: x)
    case 0
    assume "(x,y) \<in> r^\<leftrightarrow> ^^ 0" hence "x = y" by simp
    show ?case unfolding `x = y` by auto
  next
    case (Suc n)
    from `(x,y) \<in> r^\<leftrightarrow> ^^ Suc n` obtain  z where "(x,z) \<in> r^\<leftrightarrow>" and "(z,y) \<in> r^\<leftrightarrow> ^^ n"
      using rel_pow_Suc_D2 by best
    with Suc have "(z,y) \<in> r^\<down>" by simp
    from `(x,z) \<in> r^\<leftrightarrow>` show ?case
    proof
      assume "(x,z) \<in> r" with `(z,y) \<in> r^\<down>` show ?thesis by (auto intro: joinability_left_I)
    next
      assume "(x,z) \<in> r^-1"
      hence "(z,x) \<in> r^*" by simp
      from `(z,y) \<in> r^\<down>` obtain z' where "(z,z') \<in> r^*" and "(y,z') \<in> r^*" by auto
      from `CR r` and `(z,x) \<in> r^*` and `(z,z') \<in> r^*` have "(x,z') \<in> r^\<down>"
        by (rule CR_divergence_imp_joinability)
      then obtain x' where "(x,x') \<in> r^*" and "(z',x') \<in> r^*" by auto
      with `(y,z') \<in> r^*` show ?thesis by auto
    qed
  qed
next
  assume "r^\<leftrightarrow>* \<subseteq> r^\<down>" thus "CR r" unfolding CR_iff_meetability_subset_joinability
    using meetability_imp_conversion by auto
qed

lemma CR_imp_conversion_iff_joinability: assumes "CR r" shows "r^\<leftrightarrow>* = r^\<down>"
proof
  show "r^\<leftrightarrow>* \<subseteq> r^\<down>" using CR_iff_conversion_imp_joinability assms by auto
next
  show "r^\<down> \<subseteq> r^\<leftrightarrow>*" by (rule joinability_imp_conversion)
qed

lemma joinability_sym: "sym(r^\<down>)" unfolding sym_def by auto

lemma CR_joinability_left_I:
  assumes "CR r" and "(x,y) \<in> r^*" and "(x,z) \<in> r^\<down>" shows "(y,z) \<in> r^\<down>"
proof -
  from `(x,z) \<in> r^\<down>` obtain x' where "(x,x') \<in> r^*" and "(z,x') \<in> r^\<down>" by auto
  from `CR r` and `(x,x') \<in> r^*` and `(x,y) \<in> r^*` have "(x,y) \<in> r^\<down>" by auto
  hence "(y,x) \<in> r^\<down>" using joinability_sym unfolding sym_def by best
  from `CR r` have "r^\<leftrightarrow>* = r^\<down>" by (rule CR_imp_conversion_iff_joinability)
  from `(y,x) \<in> r^\<down>` and `(x,z) \<in> r^\<down>` show ?thesis using conversion_trans
    unfolding trans_def `r^\<leftrightarrow>* = r^\<down>`[symmetric] by best
qed

lemma CR_joinability_right_I:
 assumes "CR r" and "(x,y) \<in> r^\<down>" and "(y,z) \<in> r^*" shows "(x,z) \<in> r^\<down>"
proof -
  have "r^\<leftrightarrow>* = r^\<down>" by (rule CR_imp_conversion_iff_joinability[OF `CR r`])
  from `(y,z) \<in> r^*` have "(y,z) \<in> r^\<leftrightarrow>*" by (auto intro: conversion_I')
  with `(x,y) \<in> r^\<down>` show ?thesis unfolding `r^\<leftrightarrow>* = r^\<down>`[symmetric] using conversion_trans
    unfolding trans_def by fast
qed

lemma NF_not_suc: assumes "(x,y) \<in> r^*" and "x \<in> NF r" shows "x = y"
proof -
  from `x \<in> NF r` have "\<forall>y. (x,y) \<notin> r" using NF_no_step by auto
  hence "x \<notin> Domain r" unfolding Domain_def by simp
  from `(x,y) \<in> r^*` show ?thesis unfolding Not_Domain_rtrancl[OF `x \<notin> Domain r`] by simp
qed

lemma semi_complete_imp_conversion_iff_same_NF:
  assumes "semi_complete r"
  shows "((x,y) \<in> r^\<leftrightarrow>*) = (\<forall>u v. (x,u) \<in> r^! \<and> (y,v) \<in> r^! \<longrightarrow> u = v)"
proof -
  from assms have "WN r" and "CR r" unfolding semi_complete_defs by auto
  hence "r^\<leftrightarrow>* = r^\<down>" using CR_imp_conversion_iff_joinability by auto
  show ?thesis
  proof
    assume "(x,y) \<in> r^\<leftrightarrow>*"
    from `(x,y) \<in> r^\<leftrightarrow>*` have "(x,y) \<in> r^\<down>" unfolding `r^\<leftrightarrow>* = r^\<down>` .
    show "\<forall>u v. (x,u) \<in> r^! \<and> (y,v) \<in> r^! \<longrightarrow> u = v"
    proof (intro allI impI, elim conjE)
      fix u v assume "(x,u) \<in> r^!" and "(y,v) \<in> r^!"
      hence "(x,u) \<in> r^*" and "(y,v) \<in> r^*" and "u \<in> NF r" and "v \<in> NF r" by auto
      from `CR r` and `(x,u) \<in> r^*` and `(x,y) \<in> r^\<down>` have "(u,y) \<in> r^\<down>"
        by (auto intro: CR_joinability_left_I)
      hence "(y,u) \<in> r^\<down>" using joinability_sym unfolding sym_def by best
      with `(x,y) \<in> r^\<down>` have "(x,u) \<in> r^\<down>" unfolding `r^\<leftrightarrow>* = r^\<down>`[symmetric]
        using conversion_trans unfolding trans_def by best
      from `CR r` and `(x,y) \<in> r^\<down>` and `(y,v) \<in> r^*` have "(x,v) \<in> r^\<down>"
        by (auto intro: CR_joinability_right_I)
      hence "(v,x) \<in> r^\<down>" using joinability_sym unfolding sym_def by best
      with `(x,u) \<in> r^\<down>` have "(v,u) \<in> r^\<down>" unfolding `r^\<leftrightarrow>* = r^\<down>`[symmetric]
        using conversion_trans unfolding trans_def by best
      then obtain v' where "(v,v') \<in> r^*" and "(u,v') \<in> r^*" by auto
      from `(u,v') \<in> r^*` and `u \<in> NF r` have "u = v'" by (rule NF_not_suc)
      from `(v,v') \<in> r^*` and `v \<in> NF r` have "v = v'" by (rule NF_not_suc)
      thus "u = v" unfolding `u = v'` by simp
    qed
  next
    assume equal_NF:"\<forall>u v. (x,u) \<in> r^! \<and> (y,v) \<in> r^! \<longrightarrow> u = v"
    from `WN r` obtain u where "(x,u) \<in> r^!" by best
    from `WN r` obtain v where "(y,v) \<in> r^!" by best
    from `(x,u) \<in> r^!` and `(y,v) \<in> r^!` have "u = v" using equal_NF by simp
    from `(x,u) \<in> r^!` and `(y,v) \<in> r^!` have "(x,v) \<in> r^*" and "(y,v) \<in> r^*"
      unfolding `u = v` by auto
    hence "(x,v) \<in> r^\<leftrightarrow>*" and "(y,v) \<in> r^\<leftrightarrow>*" by (auto intro: conversion_I')
    from `(y,v) \<in> r^\<leftrightarrow>*` have "(v,y) \<in> r^\<leftrightarrow>*" using conversion_sym unfolding sym_def by best
    with `(x,v) \<in> r^\<leftrightarrow>*` show "(x,y) \<in> r^\<leftrightarrow>*" using conversion_trans unfolding trans_def by best
  qed
qed

lemma CR_imp_UNC: assumes "CR r" shows "UNC r"
proof - {
  fix x y assume "x \<in> NF r" and "y \<in> NF r" and "(x,y) \<in> r^\<leftrightarrow>*"
  have "r^\<leftrightarrow>* = r^\<down>" by (rule CR_imp_conversion_iff_joinability[OF assms])
  from `(x,y) \<in> r^\<leftrightarrow>*` have "(x,y) \<in> r^\<down>" unfolding `r^\<leftrightarrow>* = r^\<down>` by simp
  then obtain x' where "(x,x') \<in> r^*" and "(y,x') \<in> r^*" by best
  from `(x,x') \<in> r^*` and `x \<in> NF r` have "x = x'" by (rule NF_not_suc)
  from `(y,x') \<in> r^*` and `y \<in> NF r` have "y = x'" by (rule NF_not_suc)
  hence "x = y" unfolding `x = x'` by simp
} thus ?thesis by auto
qed

lemma WN_UNF_imp_CR: assumes "WN r" and "UNF r" shows "CR r"
proof - {
  fix x y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
  from assms obtain y' where "(y,y') \<in> r^!" unfolding WN_defs by best
  with `(x,y) \<in> r^*` have "(x,y') \<in> r^!" by (auto intro: normalizability_I')
  from assms obtain z' where "(z,z') \<in> r^!" unfolding WN_defs by best
  with `(x,z) \<in> r^*` have "(x,z') \<in> r^!" by (auto intro: normalizability_I')
  with `(x,y') \<in> r^!` have "y' = z'" using `UNF r` unfolding UNF_defs by auto
  from `(y,y') \<in> r^!` and `(z,z') \<in> r^!` have "(y,z) \<in> r^\<down>" unfolding `y' = z'` by auto
} thus ?thesis by (auto intro: CR_I')
qed

definition diamond :: "'a brel \<Rightarrow> bool" ("\<diamond>") where "\<diamond> r \<equiv> (r\<inverse> O r) \<subseteq> (r O r\<inverse>)"

lemma diamond_I[intro]: "(r\<inverse> O r) \<subseteq> (r O r\<inverse>) \<Longrightarrow> \<diamond> r" unfolding diamond_def by simp

lemma diamond_E[elim]: "\<diamond> r \<Longrightarrow> ((r\<inverse> O r) \<subseteq> (r O r\<inverse>) \<Longrightarrow> P) \<Longrightarrow> P"
  unfolding diamond_def by simp

lemma diamond_imp_semi_confluence: assumes "\<diamond> r" shows "(r\<inverse> O r^*) \<subseteq> r^\<down>"
proof (rule subsetI2)
  fix y z assume "(y,z) \<in>  r\<inverse> O r^*"
  then obtain x where "(x,y) \<in> r" and "(x,z) \<in> r^*" by best
  then obtain n where "(x,z) \<in> r^^n" using rtrancl_imp_UN_rel_pow by best
  with `(x,y) \<in> r` show "(y,z) \<in> r^\<down>"
  proof (induct n arbitrary: x z y)
    case 0 thus ?case by auto
  next
    case (Suc n)
    from `(x,z) \<in> r^^Suc n` obtain x' where "(x,x') \<in> r" and "(x',z) \<in> r^^n"
      using rel_pow_Suc_D2 by best
    with `(x,y) \<in> r` have "(y,x') \<in> (r\<inverse> O r)" by auto
    with `\<diamond> r` have "(y,x') \<in> (r O r\<inverse>)" by auto
    then obtain y' where "(x',y') \<in> r" and "(y,y') \<in> r" by best
    with Suc and `(x',z) \<in> r^^n` have "(y',z) \<in> r^\<down>" by auto
    with `(y,y') \<in> r` show ?case by (auto intro: joinability_left_I)
  qed
qed

lemma semi_confluence_imp_CR: assumes "(r\<inverse> O r^*) \<subseteq> r^\<down>" shows "CR r"
proof - {
  fix x y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
  then obtain n where "(x,z) \<in> r^^n" using rtrancl_imp_UN_rel_pow by best
  with `(x,y) \<in> r^*` have "(y,z) \<in> r^\<down>"
  proof (induct n arbitrary: x y z)
    case 0 thus ?case by auto
  next
    case (Suc n)
    from `(x,z) \<in> r^^Suc n` obtain x' where "(x,x') \<in> r" and "(x',z) \<in> r^^n"
      using rel_pow_Suc_D2 by best
    from `(x,x') \<in> r` and `(x,y) \<in> r^*` have "(x',y) \<in> (r\<inverse> O r^* )" by auto
    with assms have "(x',y) \<in> r^\<down>" by auto
    then obtain y' where "(x',y') \<in> r^*" and "(y,y') \<in> r^*" by best
    with Suc and `(x',z) \<in> r^^n` have "(y',z) \<in> r^\<down>" by simp
    then obtain u where "(z,u) \<in> r^*" and "(y',u) \<in> r^*" by best
    from `(y,y') \<in> r^*` and `(y',u) \<in> r^*` have "(y,u) \<in> r^*" by auto
    with `(z,u) \<in> r^*` show ?case by best
  qed
} thus ?thesis by (auto intro: CR_I')
qed
 
lemma diamond_imp_CR: assumes "\<diamond> r" shows "CR r"
  using assms by (rule diamond_imp_semi_confluence[THEN semi_confluence_imp_CR])

lemma diamond_imp_CR':
  assumes "\<diamond> s" and "r \<subseteq> s" and "s \<subseteq> r^*" shows "CR r"
  unfolding CR_iff_meetability_subset_joinability
proof -
  from `\<diamond> s` have "CR s" by (rule diamond_imp_CR)
  hence "s^\<up> \<subseteq> s^\<down>" unfolding CR_iff_meetability_subset_joinability by simp
  from `r \<subseteq> s` have "r^* \<subseteq> s^*" by (rule rtrancl_mono)
  from `s \<subseteq> r^*` have "s^* \<subseteq> (r^*)^*" by (rule rtrancl_mono)
  hence "s^* \<subseteq> r^*" by simp
  with `r^* \<subseteq> s^*` have "r^* = s^*" by simp
  show "r^\<up> \<subseteq> r^\<down>" unfolding meetability_def joinability_def rtrancl_converse `r^* = s^*`
    unfolding rtrancl_converse[symmetric] meetability_def[symmetric]
      joinability_def[symmetric] by (rule `s^\<up> \<subseteq> s^\<down>`)
qed

(* Only used locally *)
primrec T :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a"
where "T f x 0 = x"
    | "T f x (Suc n) = f (T f x n)" 

lemma SN_imp_minimal:
  assumes "SN r"
  shows "\<forall>Q x. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q)"
proof (rule ccontr)
  assume "\<not>(\<forall>Q x. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q))"
  then obtain Q x where "x \<in> Q" and "\<forall>z\<in>Q. \<exists>y. (z,y) \<in> r \<and> y \<in> Q" by auto
  hence "\<forall>z. \<exists>y. z \<in> Q \<longrightarrow> (z,y) \<in> r \<and> y \<in> Q" by auto
  hence "\<exists>f. \<forall>x. x \<in> Q \<longrightarrow> (x,f x) \<in> r \<and> f x \<in> Q" by (rule choice)
  then obtain f where a:"\<forall>x. x \<in> Q \<longrightarrow> (x,f x) \<in> r \<and> f x \<in> Q" (is "\<forall>x. ?P x") by best
  let ?S = "\<lambda>i. T f x i"
  have "?S 0 = x" by simp
  have "\<forall>i. (?S i,?S(Suc i)) \<in> r \<and> ?S(Suc i) \<in> Q"
  proof
    fix i show "(?S i,?S(Suc i)) \<in> r \<and> ?S(Suc i) \<in> Q" by (induct i) (auto simp:`x \<in> Q` a)
  qed
  with `?S 0 = x` have "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by fast
  with assms show False by auto
qed

lemma SN_elt_imp_elt_minimal:
  assumes "SN_elt r x"
  shows "\<forall>Q. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q)"
proof (rule ccontr)
  assume "\<not>(\<forall>Q. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q))"
  then obtain Q where "x \<in> Q" and "\<forall>z\<in>Q. \<exists>y. (z,y) \<in> r \<and> y \<in> Q" by auto
  hence "\<forall>z. \<exists>y. z \<in> Q \<longrightarrow> (z,y) \<in> r \<and> y \<in> Q" by auto
  hence "\<exists>f. \<forall>x. x \<in> Q \<longrightarrow> (x,f x) \<in> r \<and> f x \<in> Q" by (rule choice)
  then obtain f where a: "\<forall>x. x \<in> Q \<longrightarrow> (x,f x) \<in> r \<and> f x \<in> Q" (is "\<forall>x. ?P x") by best
  let ?S = "\<lambda>i. T f x i"
  have "?S 0 = x" by simp
  have "\<forall>i. (?S i,?S(Suc i)) \<in> r \<and> ?S(Suc i) \<in> Q"
  proof
    fix i show "(?S i,?S(Suc i)) \<in> r \<and> ?S(Suc i) \<in> Q" by (induct i) (auto simp:`x \<in> Q` a)
  qed
  with `?S 0 = x` have "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by fast
  with assms show False by auto
qed

hide const T

lemma minimal_imp_wf:
  assumes "\<forall>Q x. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q)"
  shows "wf(r\<inverse>)"
proof (rule ccontr)
  assume "\<not> wf(r\<inverse>)"
  hence "\<exists>P. (\<forall>x. (\<forall>y. (x,y) \<in> r \<longrightarrow> P y) \<longrightarrow> P x) \<and> (\<exists>x. \<not> P x)" unfolding wf_def by simp
  then obtain P x where suc:"\<forall>x. (\<forall>y. (x,y) \<in> r \<longrightarrow> P y) \<longrightarrow> P x" and "\<not> P x" by auto
  let ?Q = "{x. \<not> P x}"
  from `\<not> P x` have "x \<in> ?Q" by simp
  from assms have "\<forall>x. x \<in> ?Q \<longrightarrow> (\<exists>z\<in>?Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> ?Q)" by (rule allE[where x = ?Q])
  with `x \<in> ?Q` obtain z where "z \<in> ?Q" and min:" \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> ?Q" by best
  from `z \<in> ?Q` have "\<not> P z" by simp
  with suc obtain y where "(z,y) \<in> r" and "\<not> P y" by best
  hence "y \<in> ?Q" by simp
  with `(z,y) \<in> r` and min show False by simp
qed

lemmas SN_imp_wf = SN_imp_minimal[THEN minimal_imp_wf]

lemma wf_imp_SN: assumes "wf(r\<inverse>)" shows "SN r"
proof - {
  fix x
  let ?P = "\<lambda>x. \<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r))"
  from `wf(r\<inverse>)` have "?P x"
  proof induct
    case (less x)
    hence IH: "\<And>y. (x,y) \<in> r \<Longrightarrow> ?P y" by auto
    show "?P x"
    proof (rule ccontr)
      assume "\<not> ?P x"
      then obtain S where "S 0 = x" and "\<forall>i. (S i,S(Suc i)) \<in> r" by auto
      hence "(S 0,S 1) \<in> r" by auto
      with IH have "?P(S 1)" unfolding `S 0 = x` by auto
      with `\<forall>i. (S i,S(Suc i)) \<in> r` show False by auto
    qed
  qed
  hence "SN_elt r x" by auto
} thus ?thesis by auto
qed

lemma SN_iff_wf: "SN r = wf(r\<inverse>)" by (auto simp: SN_imp_wf wf_imp_SN)

subsection {* Newman's Lemma *}

lemma SN_induct:
assumes sn: "SN A" and step: "\<And>a. (\<And>b. (a,b) \<in> A \<Longrightarrow> P b) \<Longrightarrow> P a"
shows "P a"
using sn unfolding SN_iff_wf proof induct
  case (less a)
  with step show ?case by best
qed

(* The same as well-founded induction, but in the 'correct' direction. *)
lemmas SN_induct_rule = SN_induct[consumes 1, case_names IH, induct pred: SN]

lemma rtrancl_len_E[elim]: assumes "(x,y) \<in> r^*" obtains n where "(x,y) \<in> r^^n"
using rtrancl_imp_UN_rel_pow[OF assms] by best

lemma rel_pow_Suc_E2'[elim]:
assumes "(x,z) \<in> A^^Suc n" obtains y where "(x,y) \<in> A" and "(y,z) \<in> A^*"
proof -
  assume assm: "\<And>y. (x,y) \<in> A \<Longrightarrow> (y,z) \<in> A^* \<Longrightarrow> thesis"
  from rel_pow_Suc_E2[OF assms] obtain y where "(x,y) \<in> A" and "(y,z) \<in> A^^n" by auto
  hence "(y,z) \<in> A^*" using rel_pow_imp_rtrancl by auto
  from assm[OF `(x,y) \<in> A` this] show thesis .
qed

(*
The proof proceeds according to the following diagram:

x -->  s  -->* y
| WCR  |* IH   |*
t -->* u  -->* v 
|*     IH      |*
z ---------->* w
*)
lemma Newman: assumes "SN r" and "WCR r" shows "CR r"
proof
  fix x
  from `SN r` show "CR_elt r x"
  proof induct
    case (IH x) show ?case
    proof
      fix y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
      from `(x,y) \<in> r^*` obtain m where "(x,y) \<in> r^^m" ..
      from `(x,z) \<in> r^*` obtain n where "(x,z) \<in> r^^n" ..
      show "(y,z) \<in> r^\<down>"
      proof (cases n)
        case 0
        from `(x,z) \<in> r^^n` have eq: "x = z" by (simp add: 0)
        from `(x,y) \<in> r^*` show ?thesis unfolding eq ..
      next
        case (Suc n')
        from `(x,z) \<in> r^^n`[unfolded Suc] obtain t where "(x,t) \<in> r" and "(t,z) \<in> r^*" ..
        show ?thesis
        proof (cases m)
          case 0
          from `(x,y) \<in> r^^m` have eq: "x = y" by (simp add: 0)
          from `(x,z) \<in> r^*` show ?thesis unfolding eq ..
        next
          case (Suc m')
          from `(x,y) \<in> r^^m`[unfolded Suc] obtain s where "(x,s) \<in> r" and "(s,y) \<in> r^*" ..
          from `WCR r` and `(x,s) \<in> r` and `(x,t) \<in> r` have "(s,t) \<in> r^\<down>" by auto
          then obtain u where "(s,u) \<in> r^*" and "(t,u) \<in> r^*" ..
          from IH[OF `(x,s) \<in> r`] have "CR_elt r s" .
          with `(s,u) \<in> r^*` and `(s,y) \<in> r^*` have "(u,y) \<in> r^\<down>" by auto
          then obtain v where "(u,v) \<in> r^*" and "(y,v) \<in> r^*" ..
          from IH[OF `(x,t) \<in> r`] have "CR_elt r t" .
          moreover from `(t,u) \<in> r^*` and `(u,v) \<in> r^*` have "(t,v) \<in> r^*" by auto
          ultimately have "(z,v) \<in> r^\<down>" using `(t,z) \<in> r^*` by auto
          then obtain w where "(z,w) \<in> r^*" and "(v,w) \<in> r^*" ..
          from `(y,v) \<in> r^*` and `(v,w) \<in> r^*` have "(y,w) \<in> r^*" by auto
          with `(z,w) \<in> r^*` show ?thesis by auto
        qed
      qed
    qed
  qed
qed
  
lemma all_reducts_SN_imp_SN: assumes "(\<And>y. (x,y) \<in> r \<Longrightarrow> SN_elt r y)" shows "SN_elt r x"
proof (rule ccontr)
  assume "\<not> SN_elt r x"
  then obtain S where "S 0 = x" and inf:"\<forall>i. (S i,S(Suc i)) \<in> r" by auto
  let ?S = "\<lambda>i. S(Suc i)"
  from inf have "(S 0,?S 0) \<in> r" by simp
  with assms have "SN_elt r (?S 0)" unfolding `S 0 = x` by simp
  hence a: "\<not>(\<exists>S. S 0 = (?S 0) \<and> (\<forall>i. (S i,S(Suc i)) \<in> r))" by auto
  from inf have "\<forall>i. (?S i,?S(Suc i)) \<in> r" by simp
  hence "\<exists>S. S 0 = ?S 0 \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by auto
  with a show False by simp
qed

lemma SN_imp_SN_trancl: "SN R \<Longrightarrow> SN(R^+)"
  unfolding SN_iff_wf by (rule wf_converse_trancl)

lemma SN_trancl_imp_SN: assumes "SN(R^+)" shows "SN R"
proof (rule ccontr)
  assume "\<not> SN R"
  then obtain s where "\<forall>i. (s i,s(Suc i)) \<in> R" unfolding SN_defs by auto
  hence "\<forall>i. (s i,s(Suc i)) \<in> R^+" by auto
  hence "\<not> SN(R^+)" unfolding SN_defs by auto
  with assms show False by simp
qed

lemma SN_trancl_SN_conv: "SN(R^+) = SN R"
  using SN_trancl_imp_SN[of R] SN_imp_SN_trancl[of R] by blast


(* --> HOL/Relation.thy (in Isabelle) *)
lemma converse_inv_image[simp]: "(inv_image R f)^-1 = inv_image (R^-1) f"
  unfolding inv_image_def converse_def by auto

lemma SN_inv_image: "SN R \<Longrightarrow> SN(inv_image R f)" unfolding SN_iff_wf by simp

lemma SN_subset: "SN R \<Longrightarrow> R' \<subseteq> R \<Longrightarrow> SN R'" unfolding SN_defs by blast

lemma seq_imp_condensed_seq: assumes "\<forall>i. (S i,S(Suc i)) \<in> R" shows "\<forall>j>0. (S i,S(i+j)) \<in> R^^j"
proof (intro allI impI)
  fix j::nat assume "j > 0"
  thus "(S i,S(i+j)) \<in> R^^j"
  proof (induct j)
    case 0 show ?case by simp
  next
    case (Suc j) thus ?case
    proof (cases "j = 0")
      case True show ?thesis unfolding True using assms by simp
    next
      case False
      hence "j > 0" by simp
      with Suc have "(S i,S(i+j)) \<in> R^^j" by simp
      moreover from assms have "(S(i+j),S(Suc(i+j))) \<in> R" by simp
      ultimately show ?thesis by auto
    qed
  qed
qed
 
lemma SN_pow_imp_SN: assumes "SN(R^^Suc n)" shows "SN R"
proof (rule ccontr)
  assume "\<not> SN R"
  then obtain S::"nat\<Rightarrow>'a" where "\<forall>i. (S i,S(Suc i)) \<in> R" unfolding SN_defs by auto
  from seq_imp_condensed_seq[OF this]
    have step: "\<And>i. (S i,S(i+(Suc n))) \<in> R^^Suc n" by force
  let ?T = "\<lambda>i. S(i*(Suc n))"
  have "\<forall>i. (?T i,?T(Suc i)) \<in> R^^Suc n"
  proof
    fix i show "(?T i,?T(Suc i)) \<in> R^^Suc n" unfolding mult_Suc
      using step[where i="i*Suc n"] unfolding add_commute by simp
  qed
  hence "\<not> SN(R^^Suc n)" unfolding SN_defs by best
  with assms show False by simp
qed

lemma SN_pow_subset_trancl: "R^^(Suc n) \<subseteq> R^+"
proof (rule subsetI2)
  fix a b assume "(a,b) \<in> R^^Suc n" thus "(a,b) \<in> R^+"
  proof (induct n arbitrary: b)
    case 0 thus ?case by simp
  next
    case (Suc n)
    then obtain c where "(a,c) \<in> R^^Suc n" and "(c,b) \<in> R" using trancl_power by auto
    with Suc have "(a,c)\<in> R^+" by simp
    with `(c,b) \<in> R` show ?case by auto
  qed
qed

lemma SN_imp_SN_pow: assumes "SN R" shows "SN(R^^Suc n)"
  using SN_subset[where R="R^+",OF SN_imp_SN_trancl[OF assms] SN_pow_subset_trancl] by simp
  
(* needed in HOL/Wellfounded.thy *)
lemma SN_pow: "SN R \<longleftrightarrow> SN(R ^^ Suc n)"
  by (rule iffI,rule SN_imp_SN_pow,assumption,rule SN_pow_imp_SN,assumption)

lemma SN_elt_imp_SN_elt_trancl: assumes "SN_elt R t" shows "SN_elt (R^+) t"
using assms proof (rule contrapos_pp)
  let ?R = "SN_rel R R"
  assume "\<not> SN_elt (R^+) t"
  then obtain S :: "nat \<Rightarrow> 'a" where "S 0 = t" and S: "\<forall>i. (S i,S(Suc i)) \<in> R^+"
    unfolding SN_elt_def by auto
  have "SN ?R" by (rule SN_SN_rel)
  hence "SN(?R^+)" by (rule SN_imp_SN_trancl)
  have "\<forall>i. (t,S i) \<in> R^*"
  proof
    fix i show "(t,S i) \<in> R^*"
    proof (induct i)
      case 0 show ?case unfolding `S 0 = t` by simp
    next
      case (Suc i)
      from S have "(S i,S(Suc i)) \<in> R^+" by simp
      with Suc show ?case by auto
    qed
  qed
  with assms have "\<forall>i. SN_elt R (S i)" using rtrancl_step_preserves_SN[where s = t and R = R]
    by auto
  with S have "\<forall>i. (S i,S(Suc i)) \<in> ?R^+" unfolding SN_rel_trancl_simp unfolding SN_rel_def
    by auto
  hence "\<not> SN_elt (?R^+) t" unfolding `S 0 = t`[symmetric] unfolding SN_defs by auto
  with `SN(?R^+)` have "False" unfolding SN_defs by simp
  thus "\<not> SN_elt R t" by simp
qed


definition "restrict" :: "'a brel \<Rightarrow> 'a set \<Rightarrow> 'a brel"
where "restrict r S \<equiv> {(x,y). x \<in> S \<and> y \<in> S \<and> (x,y) \<in> r}"

lemma SN_imp_restrict_SN_elt:
  assumes "SN_elt r x" shows "SN_elt (restrict r S) x" (is "SN_elt ?r x")
proof (rule ccontr)
  assume "\<not> SN_elt ?r x"
  hence "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> ?r)" by auto
  hence "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" unfolding restrict_def by auto
  with `SN_elt r x` show False by auto
qed

lemma SN_imp_restrict_SN: "SN r \<Longrightarrow> SN(restrict r S)"
  using SN_imp_restrict_SN_elt unfolding SN_def by best

lemma restrict_rtrancl: "(restrict r S)^* \<subseteq> r^*" (is "?r^* \<subseteq> r^*")
proof - {
  fix x y assume "(x,y) \<in> ?r^*" hence "(x,y) \<in> r^*" unfolding restrict_def by induct auto
} thus ?thesis by auto
qed

text {* ALN, Exercise~1.10.a *}
lemma SN_imp_SN_reduct: assumes "SN_elt r x" and "(x,y) \<in> r^*" shows "SN_elt r y"
proof - {
  assume a:"\<exists>S. S 0 = y \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)"
  from `(x,y) \<in> r^*` obtain n where "(x,y) \<in> r^^n" using rtrancl_imp_UN_rel_pow by best
  hence "\<not> SN_elt r x"
  proof (induct n arbitrary: x)
    case 0 hence "x = y" by simp
    show ?case using a unfolding `x = y` by auto
  next
    case (Suc n)
    from `(x,y) \<in> r^^Suc n` obtain x' where "(x,x') \<in> r" and "(x',y) \<in> r^^n"
      using rel_pow_Suc_D2 by best
    with Suc have "\<not> SN_elt r x'" by simp
    then obtain S where "S 0 = x'" and "\<forall>i. (S i,S(Suc i)) \<in> r" by auto
    let ?S = "\<lambda>i. case i of 0 \<Rightarrow> x | Suc i \<Rightarrow> S i"
    have "?S 0 = x" by simp
    have "\<forall>i. (?S i,?S(Suc i)) \<in> r"
    proof
      fix i show "(?S i,?S(Suc i)) \<in> r"
        by (induct i) (auto simp: `S 0 = x'` `(x,x') \<in> r` `\<forall>i. (S i,S(Suc i)) \<in> r`)
    qed
    with `?S 0 = x` have "\<exists>S. S 0 = x \<and> (\<forall>i. (S i,S(Suc i)) \<in> r)" by best
    thus ?case by auto
   qed
} with `SN_elt r x` show ?thesis by auto
qed

lemma WCR_SN_elt_imp_CR_elt: assumes "WCR r" and "SN_elt r x" shows "CR_elt r x"
proof -
  let ?S = "{y. (x,y) \<in> r^*}"
  let ?r = "restrict r ?S"
  have "\<forall>x. SN_elt ?r x"
  proof
    fix y have "y \<notin> ?S \<or> y \<in> ?S" by simp 
    thus "SN_elt ?r y"
    proof
      assume "y \<notin> ?S" thus ?thesis unfolding restrict_def by auto
    next
      assume "y \<in> ?S"
      hence "(x,y) \<in> r^*" by simp
      with `SN_elt r x` have "SN_elt r y" by (rule SN_imp_SN_reduct)
      thus ?thesis by (rule SN_imp_restrict_SN_elt)
    qed
  qed
  hence "SN ?r" by auto
  {
    fix x y assume "(x,y) \<in> r^*" and "x \<in> ?S" and "y \<in> ?S"
    then obtain n where "(x,y) \<in> r^^n" and "x \<in> ?S" and "y \<in> ?S"
      using rtrancl_imp_UN_rel_pow by best
    hence "(x,y) \<in> ?r^*"
    proof (induct n arbitrary: x y)
      case 0 thus ?case by simp
    next
      case (Suc n)
      from `(x,y) \<in> r^^Suc n` obtain x' where "(x,x') \<in> r" and "(x',y) \<in> r^^n"
        using rel_pow_Suc_D2 by best
      hence "(x,x') \<in> r^*" by simp
      with `x \<in> ?S` have "x' \<in> ?S" by auto
      with Suc and `(x',y) \<in> r^^n` have "(x',y) \<in> ?r^*" by simp
      from `(x,x') \<in> r` and `x \<in> ?S` and `x' \<in> ?S` have "(x,x') \<in> ?r"
        unfolding restrict_def by simp
      with `(x',y) \<in> ?r^*` show ?case by simp
    qed
  }
  hence a:"\<forall>x y. (x,y) \<in> r^* \<and> x \<in> ?S \<and> y \<in> ?S \<longrightarrow> (x,y) \<in> ?r^*" by simp
  {
    fix x' y z assume "(x',y) \<in> ?r" and "(x',z) \<in> ?r"
    hence "x' \<in> ?S" and "y \<in> ?S" and "z \<in> ?S" and "(x',y) \<in> r" and "(x',z) \<in> r"
      unfolding restrict_def by auto
    with `WCR r` have "(y,z) \<in> r^\<down>" by force
    then obtain u where "(y,u) \<in> r^*" and "(z,u) \<in> r^*" by auto
    from `x' \<in> ?S` have "(x,x') \<in> r^*" by simp
    from `(x',y) \<in> r` have "(x',y) \<in> r^*" by auto
    with `(y,u) \<in> r^*` have "(x',u) \<in> r^*" by auto
    with `(x,x') \<in> r^*` have "(x,u) \<in> r^*" by simp
    hence "u \<in> ?S" by simp
    from `y \<in> ?S` and `u \<in> ?S` and `(y,u) \<in> r^*` have "(y,u) \<in> ?r^*" using a by auto
    from `z \<in> ?S` and `u \<in> ?S` and `(z,u) \<in> r^*` have "(z,u) \<in> ?r^*" using a by auto
    with `(y,u) \<in> ?r^*` have "(y,z) \<in> ?r^\<down>" by auto
  }
  hence "WCR ?r" by (auto intro: WCR_I')
  have "CR ?r" using Newman[OF `SN ?r` `WCR ?r`] by simp
  {
    fix y z assume "(x,y) \<in> r^*" and "(x,z) \<in> r^*"
    hence "y \<in> ?S" and "z \<in> ?S" by auto
    have "x \<in> ?S" by simp
    from a and `(x,y) \<in> r^*` and `x \<in> ?S` and `y \<in> ?S` have "(x,y) \<in> ?r^*" by simp
    from a and `(x,z) \<in> r^*` and `x \<in> ?S` and `z \<in> ?S` have "(x,z) \<in> ?r^*" by simp
    with `CR ?r` and `(x,y) \<in> ?r^*` have "(y,z) \<in> ?r^\<down>" by force
    then obtain u where "(y,u) \<in> ?r^*" and "(z,u) \<in> ?r^*" by best
    hence "(y,u) \<in> r^*" and "(z,u) \<in> r^*" using restrict_rtrancl by auto
    hence "(y,z) \<in> r^\<down>" by auto
  }
  thus ?thesis by auto
qed

lemma rtrancl_imp_seq:
  assumes "(x,y) \<in> r^*"
  shows "\<exists>S n. S 0 = x \<and> S n = y \<and> (\<forall>i<n. (S i,S(Suc i)) \<in> r)"
using assms proof (induct)
   case base thus ?case by auto
next
  case (step y z)
  then obtain S n where "S 0 = x" and "S n = y" and seq: "\<forall>i<n. (S i,S(Suc i)) \<in> r" by auto
  let ?m = "Suc n"
  let ?S = "\<lambda>i. if i = ?m then z else S i"
  have "?S ?m = z" by simp
  from `S 0 = x` have "?S 0 = x" by simp
  from seq have seq': "\<forall>i<n. (?S i,?S(Suc i)) \<in> r" by auto
  with `S n = y` and `(y,z) \<in> r` have "\<forall>i<?m. (?S i,?S(Suc i)) \<in> r" by auto
  with `?S 0 = x` and `?S ?m = z` show ?case by best
qed

lemma SN_imp_normalizability: assumes "SN_elt r x" shows "\<exists>y. (x,y) \<in> r^!"
proof (rule ccontr)
  assume "\<not>(\<exists>y. (x,y) \<in> r^!)"
  hence A: "\<forall>y. (x,y) \<in> r^* \<longrightarrow> y \<notin> NF r" by auto
  hence "x \<notin> NF r" by auto
  let ?Q = "{z. (x,z) \<in> r^* \<and> z \<notin> NF r}"
  have "x \<in> ?Q" using `x \<notin> NF r` by simp
  have "\<forall>z\<in>?Q. \<exists>y. (z,y) \<in> r \<and> y \<in> ?Q"
  proof
    fix z assume "z \<in> ?Q"
    hence "(x,z) \<in> r^*" and "z \<notin> NF r" by auto
    then obtain u where "(z,u) \<in> r" by auto
    with `(x,z) \<in> r^*` have "(x,u) \<in> r^*" by simp
    with A have "u \<notin> NF r" by simp
    with `(z,u) \<in> r` and `(x,u) \<in> r^*`
    show "\<exists>y. (z,y) \<in> r \<and> y \<in> ?Q" by auto
  qed
  with `x \<in> ?Q` have "x \<in> ?Q \<and> (\<forall>z\<in>?Q. \<exists>y. (z,y) \<in> r \<and> y \<in> ?Q)" by auto
  hence "\<exists>Q. x \<in> Q \<and> (\<forall>z\<in>Q. \<exists>y. (z,y) \<in> r \<and> y \<in> Q)" by (rule exI[where x = "?Q"])
  hence "\<not>(\<forall>Q. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q))" by simp
  with SN_elt_imp_elt_minimal[of r x] have "\<not> SN_elt r x" by auto
  with assms show False by simp
qed


subsection {* Commutation *}

definition commute :: "'a brel \<Rightarrow> 'a brel \<Rightarrow> bool"
where "commute r s \<equiv> ((r\<inverse>)^* O s^*) \<subseteq> (s^* O (r\<inverse>)^*)"

lemma CR_iff_self_commute: "CR r = commute r r"
  unfolding commute_def CR_iff_meetability_subset_joinability meetability_def joinability_def
  by simp

(* --> belongs somewhere else *)
lemma rtrancl_imp_rtrancl_UN: 
  assumes "(x,y) \<in> r^*" and "r \<in> I"
  shows "(x,y) \<in> (\<Union>r\<in>I. r)^*" (is "(x,y) \<in> ?r^*")
using assms proof induct
  case base thus ?case by simp
next
  case (step y z)
  hence "(x,y) \<in> ?r^*" by simp
  from `(y,z) \<in> r` and `r \<in> I` have "(y,z) \<in> ?r^*" by auto
  with `(x,y) \<in> ?r^*` show ?case by auto
qed

definition quasi_commute :: "'a brel \<Rightarrow> 'a brel \<Rightarrow> bool"
where "quasi_commute r s \<equiv> (s O r) \<subseteq> r O (r \<union> s)^*"

lemma rtrancl_union_subset_rtrancl_union_trancl: "(r \<union> s^+)^* = (r \<union> s)^*"
proof
  show "(r \<union> s^+)^* \<subseteq> (r \<union> s)^*"
  proof (rule subsetI2)
    fix x y assume "(x,y) \<in> (r \<union> s^+)^*"
    thus "(x,y) \<in> (r \<union> s)^*"
    proof (induct)
      case base thus ?case by auto
    next
      case (step y z)
      hence "(y,z) \<in> r \<or> (y,z) \<in> s^+" by auto
      hence "(y,z) \<in> (r \<union> s)^*"
      proof
        assume "(y,z) \<in> r" thus ?thesis by auto
      next
        assume "(y,z) \<in> s^+"
        hence "(y,z) \<in> s^*" by auto
        hence "(y,z) \<in> r^* \<union> s^*" by auto
        thus ?thesis using rtrancl_Un_subset by auto
      qed
      with `(x,y) \<in> (r \<union> s)^*` show ?case by simp
    qed
  qed
next
  show "(r \<union> s)^* \<subseteq> (r \<union> s^+)^*"
  proof (rule subsetI2)
    fix x y assume "(x,y) \<in> (r \<union> s)^*"
    thus "(x,y) \<in> (r \<union> s^+)^*"
    proof (induct)
      case base thus ?case by auto
    next
      case (step y z)
      hence "(y,z) \<in> (r \<union> s^+)^*" by auto
      with `(x,y) \<in> (r \<union> s^+)^*` show ?case by auto
    qed
  qed
qed

lemma qc_imp_qc_trancl: assumes "quasi_commute r s" shows "quasi_commute r (s^+)"
unfolding quasi_commute_def proof (rule subsetI2)
  fix x z assume "(x,z) \<in> s^+ O r"
  then obtain y where "(x,y) \<in> s^+" and "(y,z) \<in> r" by best
  thus "(x,z) \<in> r O (r \<union> s^+)^*"
  proof (induct arbitrary: z)
    case (base y)
    hence "(x,z) \<in> (s O r)" by auto
    with assms have "(x,z) \<in> r O (r \<union> s)^*" unfolding quasi_commute_def by auto
    thus ?case using rtrancl_union_subset_rtrancl_union_trancl by auto
  next
    case (step a b)
    hence "(a,z) \<in> (s O r)" by auto
    with assms have "(a,z) \<in> r O (r \<union> s)^*" unfolding quasi_commute_def by auto
    then obtain u where "(a,u) \<in> r" and "(u,z) \<in> (r \<union> s)^*" by best
    hence "(u,z) \<in> (r \<union> s^+)^*" using rtrancl_union_subset_rtrancl_union_trancl by auto
    from `(a,u) \<in> r` and step have "(x,u) \<in> r O (r \<union> s^+)^*" by auto
    then obtain v where "(x,v) \<in> r" and "(v,u) \<in> (r \<union> s^+)^*" by best
    with `(u,z) \<in> (r \<union> s^+)^*` have "(v,z) \<in> (r \<union> s^+)^*" by auto
    with `(x,v) \<in> r` show ?case by auto
  qed
qed

lemma not_SN_reduct_imp_not_SN:
  assumes "\<not> SN_elt r y" and "(x,y) \<in> r^*"
  shows "\<not> SN_elt r x"
  using SN_imp_SN_reduct[where r = r and x = x and y = y] assms by auto

lemma elt_of_nontermin_seq_is_nontermin:
 assumes "\<And>i. (S i,S(Suc i)) \<in> r" shows "\<And>i. \<not> SN_elt r (S i)"
proof -
  fix i
  let ?S = "\<lambda>j. S(i + j)"
  have "?S 0 = S i" by simp
  have "\<forall>i. (?S i,?S(Suc i)) \<in> r" using assms by auto
  with `?S 0 = S i` have "?S 0 = S i \<and> (\<forall>i. (?S i,?S(Suc i)) \<in> r)" by auto
  hence "\<exists>T. T 0 = S i \<and> (\<forall>i. (T i,T(Suc i)) \<in> r)" by best
  thus "\<not> SN_elt r (S i)" unfolding SN_elt_def by auto
qed

lemma quasi_commute_imp_SN:
  assumes "SN r" and "SN s" and "quasi_commute r s"
  shows "SN(r \<union> s)"
proof -
  have "quasi_commute r (s^+)" by (rule qc_imp_qc_trancl[OF `quasi_commute r s`])
  let ?B = "{a. \<not> SN_elt (r \<union> s) a}"
  {
    assume "\<not> SN(r \<union> s)"
    then obtain a where "a \<in> ?B" unfolding SN_defs by best
    from `SN r` have "\<forall>Q x. x \<in> Q \<longrightarrow> (\<exists>z\<in>Q. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> Q)"
      by (rule SN_imp_minimal)
    hence "\<forall>x. x \<in> ?B \<longrightarrow> (\<exists>z\<in>?B. \<forall>y. (z,y) \<in> r \<longrightarrow> y \<notin> ?B)" by (rule spec[where x = ?B])
    with `a \<in> ?B` obtain b where "b \<in> ?B" and min: "\<forall>y. (b,y) \<in> r \<longrightarrow> y \<notin> ?B" by auto
    from `b \<in> ?B` obtain S where "S 0 = b" and
      seq: "\<forall>i. (S i,S(Suc i)) \<in> r \<union> s" unfolding SN_elt_def by auto
    let ?S = "\<lambda>i. S(Suc i)"
    have "?S 0 = S 1" by simp
    from seq have "\<forall>i. (?S i,?S(Suc i)) \<in> r \<union> s" by auto
    with `?S 0 = S 1` have "\<not> SN_elt (r \<union> s) (S 1)" unfolding SN_elt_def by auto
    from `S 0 = b` and seq have "(b,S 1) \<in> r \<union> s" by auto
    with min and `\<not> SN_elt (r \<union> s) (S 1)` have "(b,S 1) \<in> s" by auto
    let ?i = "LEAST i. (S i,S(Suc i)) \<notin> s"
    {
      assume "\<forall>i. (S i,S(Suc i)) \<in> s"
      with `S 0 = b` have "\<not> SN_elt s b" unfolding SN_elt_def by auto
      with `SN s` have False unfolding SN_def by auto
    }
    hence ex: "\<exists>i. (S i,S(Suc i)) \<notin> s" by auto
    hence "(S ?i,S(Suc ?i)) \<notin> s" by (rule LeastI_ex)
    with seq have "(S ?i,S(Suc ?i)) \<in> r" by auto
    have ini: "\<forall>i<?i. (S i,S(Suc i)) \<in> s" using not_less_Least by auto
    {
      fix i assume "i < ?i" hence "(b,S(Suc i)) \<in> s^+"
      proof (induct i)
        case 0 thus ?case using `(b,S 1) \<in> s` and `S 0 = b` by auto
      next
        case (Suc k)
      hence "(b,S(Suc k)) \<in> s^+" and "Suc k < ?i" by auto
      with `\<forall>i<?i. (S i,S(Suc i)) \<in> s` have "(S(Suc k),S(Suc(Suc k))) \<in> s" by fast
      with `(b,S(Suc k)) \<in> s^+` show ?case by auto
    qed
    }
    hence pref: "\<forall>i<?i. (b,S(Suc i)) \<in> s^+" by auto
    from `(b,S 1) \<in> s` and `S 0 = b` have "(S 0,S(Suc 0)) \<in> s" by auto
    {
      assume "?i = 0"
      from ex have "(S ?i,S(Suc ?i)) \<notin> s" by (rule LeastI_ex)
      with `(S 0,S(Suc 0)) \<in> s` have False unfolding `?i = 0` by simp
    }
    hence "0 < ?i" by auto
    then obtain j where "?i =  Suc j" unfolding gr0_conv_Suc by best
    with ini have "(S(?i-Suc 0),S(Suc(?i-Suc 0))) \<in> s" by auto
    with pref have "(b,S(Suc j)) \<in> s^+" unfolding `?i = Suc j` by auto
    hence "(b,S ?i) \<in> s^+" unfolding `?i = Suc j` by auto
    with `(S ?i,S(Suc ?i)) \<in> r` have "(b,S(Suc ?i)) \<in> (s^+ O r)" by auto
    with `quasi_commute r (s^+)` have "(b,S(Suc ?i)) \<in> r O (r \<union> s^+)^*"
      unfolding quasi_commute_def by auto
    then obtain c where "(b,c) \<in> r" and "(c,S(Suc ?i)) \<in> (r \<union> s^+)^*" by best
    from `(b,c) \<in> r` have "(b,c) \<in> (r \<union> s)^*" by auto
    from elt_of_nontermin_seq_is_nontermin[where r = "r \<union> s" and S = S]
      and seq have "\<not> SN_elt (r \<union> s) (S(Suc ?i))" by auto
    from `(c,S(Suc ?i)) \<in> (r \<union> s^+)^*` have "(c,S(Suc ?i)) \<in> (r \<union> s)^*"
      unfolding rtrancl_union_subset_rtrancl_union_trancl by auto
    with not_SN_reduct_imp_not_SN[where r = "r \<union> s" and x = c]
      and `\<not> SN_elt (r \<union> s) (S(Suc ?i))` have "\<not> SN_elt (r \<union> s) c" by auto
    hence "c \<in> ?B" by simp
    with `(b,c) \<in> r` and min have False by auto
  }
  thus ?thesis by auto
qed

lemma non_strict_into_strict:
  assumes compat: "NS O S \<subseteq> S"
  and steps: "(s,t) \<in> (NS^*) O S"
  shows "(s,t) \<in> S"
using steps proof
  fix x u z
  assume "(s,t) = (x,z)" and "(x,u) \<in> NS^*" and "(u,z) \<in> S"
  hence "(s,u) \<in> NS^*" and "(u,t) \<in> S" by auto
  thus ?thesis
  proof (induct rule:rtrancl.induct)
    case (rtrancl_refl x) thus ?case .
  next
    case (rtrancl_into_rtrancl a b c)
    with compat show ?case by auto
  qed
qed

lemma comp_trancl: assumes "R O S \<subseteq> S" shows "R O S^+ \<subseteq> S^+"
proof (rule subsetI2)
  fix w z assume "(w,z) \<in> R O S^+"
  then obtain x where R_step: "(w,x) \<in> R" and S_seq: "(x,z) \<in> S^+" by best
  from tranclD[OF S_seq] obtain y where S_step: "(x,y) \<in> S" and S_seq': "(y,z) \<in> S^*" by auto
  from R_step and S_step have "(w,y) \<in> R O S" by auto
  with assms have "(w,y) \<in> S" by auto
  with S_seq' show "(w,z) \<in> S^+" by simp
qed

lemma comp_rtrancl_into_trancl:
  assumes comp: "R O S \<subseteq> S"
    and seq: "(s,t) \<in> (R \<union> S)^* O S"
  shows "(s,t) \<in> S^+"
using seq proof
  fix x u z
  assume "(s,t) = (x,z)" and "(x,u) \<in> (R \<union> S)^*" and "(u,z) \<in> S"
  hence "(s,u) \<in> (R \<union> S)^*" and "(u,t) \<in> S^+" by auto
  thus ?thesis
  proof (induct rule: rtrancl.induct)
    case (rtrancl_refl x) thus ?case .
  next
    case (rtrancl_into_rtrancl a b c)
    hence "(b,c) \<in> R \<union> S" by simp
    thus ?case
    proof
      assume "(b,c) \<in> S"
      with rtrancl_into_rtrancl
      have "(b,t) \<in> S^+" by simp
      with rtrancl_into_rtrancl show ?thesis by simp
    next
      assume "(b,c) \<in> R"
      with comp_trancl[OF comp] rtrancl_into_rtrancl
      show ?thesis by auto
    qed
  qed
qed

lemma steps_imp_seq:
fixes i j :: nat
assumes seq: "\<forall>i. (s i, s (Suc i)) \<in> A"
  and "j \<ge> i"
shows "(s i, s j) \<in> A^*"
using `j \<ge> i` proof (induct j)
  case 0 thus ?case by simp
next
  case (Suc j)
  show ?case
  proof (cases "i = Suc j")
    case True show ?thesis by (simp add: True)
  next
    case False
    with Suc have "(s i, s j) \<in> A^*" by simp
    moreover from seq have "(s j, s (Suc j)) \<in> A^*" by auto
    ultimately show ?thesis by simp
  qed
qed

lemma trancl_union_right: fixes r::"'a brel" shows "r^+ \<subseteq> (s \<union> r)^+"
proof (rule subsetI2)
  fix x y assume "(x,y) \<in> r^+" thus "(x,y) \<in> (s \<union> r)^+"
  proof (induct)
    case base thus ?case by auto
  next
    case (step a b)
    hence "(a,b) \<in> (s \<union> r)^+" by auto
    with `(x,a) \<in> (s \<union> r)^+` show ?case by auto
  qed
qed

lemma SN_rel_subset: "SN_rel R S \<subseteq> R"
proof (rule subsetI2)
  fix a b assume "(a,b) \<in> SN_rel R S" thus "(a,b) \<in> R" unfolding SN_rel_def by simp
qed

lemma union_seq_SN_elt_imp_first_step:
  assumes "\<forall>i. (t i, t (Suc i)) \<in> (R \<union> S)" and "SN_elt S (t 0)"
  shows "\<exists>i. (t i, t (Suc i)) \<in> R \<and> (\<forall>j<i. (t j, t (Suc j)) \<in> S \<and> (t j, t (Suc j)) \<notin> R)"
proof -
  from `SN_elt S (t 0)` obtain i where "(t i, t (Suc i)) \<notin> S" by blast
  with assms have "(t i, t (Suc i)) \<in> R" (is "?P i") by auto
  let ?i = "Least ?P"
  from `?P i` have "?P ?i" by (rule LeastI)
  have "\<forall>j<?i. (t j, t (Suc j)) \<notin> R" using not_less_Least by auto
  moreover with assms have "\<forall>j<?i. (t j, t (Suc j)) \<in> S" by best
  ultimately have "\<forall>j<?i. (t j, t (Suc j)) \<in> S \<and> (t j, t (Suc j)) \<notin> R" by best
  with `?P ?i` show ?thesis by best
qed

fun idx
where "idx f 0 = 0"
    | "idx f (Suc i) = Suc(f(idx f i))"

lemma non_strict_ending:
  assumes seq: "\<forall>i. (t i,t(Suc i)) \<in> R \<union> S"
    and comp: "R O S \<subseteq> S"
    and SN: "SN_elt S (t 0)"
  shows "\<exists>j. \<forall>i\<ge>j. (t i,t(Suc i)) \<in> R - S" (is ?thesis)
proof (rule ccontr)
  assume "\<not> ?thesis"
  with seq have "\<forall>j.\<exists>i. i\<ge>j \<and> (t i,t(Suc i)) \<in> S" by blast
  from choice[OF this] obtain f where S_steps: "\<forall>i. i\<le>f i \<and> (t(f i),t(Suc(f i))) \<in> S" ..
  let ?t = "\<lambda>i. t(idx f i)"
  have S_seq: "\<forall>i. (t i,t(Suc(f i))) \<in> S^+"
  proof
    fix i
    from S_steps have leq: "i\<le>f i" and step: "(t(f i),t(Suc(f i))) \<in> S" by auto
    from steps_imp_seq[OF seq leq] have "(t i,t(f i)) \<in> (R \<union> S)^*" by simp
    with step have "(t i,t(Suc(f i))) \<in> (R \<union> S)^* O S" by auto
    from comp_rtrancl_into_trancl[OF comp this] show "(t i,t(Suc(f i))) \<in> S^+" .
  qed
  hence "\<forall>i. (?t i,?t(Suc i)) \<in> S^+" by simp
  moreover have "SN_elt (S^+) (?t 0)" using SN_elt_imp_SN_elt_trancl[OF SN] by simp
  ultimately show False unfolding SN_defs by best
qed

hide const idx

lemma SN_imp_SN_subset:
  assumes "SN_elt R' x" and "rR' \<subseteq> R'" shows "SN_elt rR' x" unfolding SN_elt_def
proof-
  from assms have noS: "\<not>(\<exists>S. S 0 = x \<and> (\<forall>i. (S i, S (Suc i)) \<in> R'))" unfolding SN_elt_def
    by simp
  show " \<not> (\<exists>S. S 0 = x \<and> (\<forall>i. (S i, S (Suc i)) \<in> rR'))"
  proof(rule notI)
    assume "\<exists>S. S 0 = x \<and> (\<forall>i. (S i, S (Suc i)) \<in> rR')"
    then obtain S where "S 0 = x" and "\<forall>i. (S i, S (Suc i)) \<in> rR'" by auto
    with `rR' \<subseteq> R'` have "\<forall>i. (S i, S (Suc i)) \<in> R'" by auto
    with `S 0 = x` have "\<exists> S. (S 0 = x) \<and> (\<forall>i. (S i, S (Suc i)) \<in> R')" by auto
    with noS show False ..
  qed
qed


lemma trancl_to_rtrancl: "R^+ = R^* O R"
by (fact trancl_unfold_right) (* FIXME *)

lemma rtrancl_imp_rel_pow': "(x,y) \<in> R^* \<Longrightarrow> \<exists>n. (x,y) \<in> ((R::'a brel) ^^ n)"
proof (induct rule: rtrancl_induct)
  case base thus ?case using rel_pow_0_I by best
next
  case (step y z)
  then obtain n where "(x,y) \<in> (R ^^ n)" by best
  with step have "(x,z) \<in> (R ^^ (Suc n))" using rel_pow_Suc_I by auto
  thus ?case by best
qed

lemma compat_tr_compat: assumes "NS O S \<subseteq> S" shows "NS^* O S \<subseteq> S"
  using non_strict_into_strict[where S = S and NS = NS] assms by blast

lemma right_comp_S[simp]:
  assumes "(x, y) \<in> S O (S O S^* O NS^* \<union> NS^*)"
  shows "(x, y) \<in> (S O S^* O NS^*)"
proof-
  from assms have "(x, y) \<in> (S O S O S^* O NS^*) \<union> (S O NS^*)" by auto
  hence  xy:"(x, y) \<in> (S O (S O S^*) O NS^*) \<union> (S O NS^*)" by auto
  have "S O S^* \<subseteq> S^*" by auto
  with xy have "(x, y) \<in> (S O S^* O NS^*) \<union> (S O NS^*)" by auto
  then show "(x, y) \<in> (S O S^* O NS^*)" by auto
qed

lemma NS_comp_S_is_SN:
  assumes SN: "SN S"
  and compat: "NS O S \<subseteq> S" 
  shows "SN (S O S^* O NS^*)"
proof
  fix x::'a
  from compat compat_tr_compat have tr_compat: "NS^* O S \<subseteq> S" by blast
  show "SN_elt (S O S^* O NS^*) x"
  proof (rule ccontr)
    assume "\<not> SN_elt (S O S^* O NS^*) x"
    hence F: "\<exists>F. F 0 = x \<and> (\<forall>i. (F i,F(Suc i)) \<in> (S O S^* O NS^*))"
      unfolding SN_elt_def by auto
    from F obtain F where "F 0 = x" and alli: "\<forall> i. (F i,F(Suc i)) \<in> (S O S^* O NS^*)" by auto
    have "\<forall> i. (\<exists> y z. (F i, y)  \<in> S \<and> (y, z)  \<in> S^* \<and> (z, F (Suc i)) \<in> NS^*)"
    proof
      fix i
      from alli have "(F i,F (Suc i)) \<in> (S O S^* O NS^*)" by auto
      thus "\<exists> y z. (F i, y)  \<in> S \<and> (y, z)  \<in> S^* \<and> (z, F (Suc i)) \<in> NS^*"
        unfolding rel_comp_def using mem_Collect_eq by auto
    qed
    hence "\<exists> f. (\<forall> i. (\<exists> z. (F i, f i)  \<in> S \<and> ((f i, z)  \<in> S^*) \<and>(z, F (Suc i)) \<in> NS^*))"
      by (rule choice)
    then obtain f
      where "\<forall> i. (\<exists> z. (F i, f i)  \<in> S \<and> ((f i, z)  \<in> S^*) \<and>(z, F (Suc i)) \<in> NS^*)" ..
    hence "\<exists> g. \<forall> i. (F i, f i)  \<in> S \<and> (f i, g i)  \<in> S^* \<and> (g i, F (Suc i)) \<in> NS^*"
      by (rule choice)
    then obtain g where "\<forall> i. (F i, f i)  \<in> S \<and> (f i, g i)  \<in> S^* \<and> (g i, F (Suc i)) \<in> NS^*" ..
    hence "\<forall> i. (f i, g i)  \<in> S^* \<and> (g i, F (Suc i)) \<in> NS^* \<and> (F (Suc i), f (Suc i))  \<in> S"
      by auto
    hence "\<forall> i. (f i, g i)  \<in> S^* \<and> (g i, f (Suc i))  \<in> S" unfolding rel_comp_def
      using tr_compat by auto
    hence all:"\<forall> i. (f i, g i)  \<in> S^* \<and> (g i, f (Suc i))  \<in> S^+" by auto
    have "\<forall> i. (f i, f (Suc i))  \<in> S^+"
    proof
      fix i
      from all have "(f i, g i)  \<in> S^* \<and> (g i, f (Suc i))  \<in> S^+" ..
      thus "(f i, f (Suc i))  \<in> S^+" using transitive_closure_trans by auto
    qed
    hence "\<exists> x. ((f 0) = x) \<and> (\<forall>i. (f i, f (Suc i)) \<in> S^+)" by auto
    then obtain x where "((f 0) = x) \<and> (\<forall>i. (f i, f (Suc i)) \<in> S^+)" by auto
    hence "\<exists> f. ((f 0) = x) \<and> (\<forall>i. (f i, f (Suc i)) \<in> S^+)" by auto
    hence "\<not> SN_elt (S^+) x" by auto
    hence "\<not> SN (S^+)" by auto
    hence wfSconv:"\<not> wf ((S^+)\<inverse>)" using SN_iff_wf by auto
    from SN have "wf (S\<inverse>)" using SN_imp_wf[where?r=S] by simp
    with wf_converse_trancl wfSconv show False by auto
  qed
qed

lemma split_NS_S_union:
  assumes compat: "NS O S \<subseteq> S"
   and steps: "(x, y) \<in> (NS \<union> S)^*"
  shows "(x, y) \<in> S O S^* O NS^* \<union> NS^*"
proof-
  from steps have "\<exists> n. (x, y) \<in> (NS \<union> S)^^n" using rtrancl_imp_rel_pow'[where ?R="NS \<union> S"] by auto
  then obtain n where "(x, y) \<in> (NS \<union> S)^^n" by auto
  thus "(x, y) \<in>  S O S^* O NS^* \<union> NS^*"
  proof (induct n arbitrary: x, simp)
    case (Suc m)
    assume "(x, y) \<in> (NS \<union> S)^^(Suc m)"
    hence "\<exists> z. (x, z) \<in> (NS \<union> S) \<and> (z, y) \<in> (NS \<union> S)^^m" using rel_pow_Suc_D2[where ?R="NS \<union> S"] by auto
    then obtain z where xz:"(x, z) \<in> (NS \<union> S)" and zy:"(z, y) \<in> (NS \<union> S)^^m" by auto
    with Suc have zy:"(z, y) \<in>  S O S^* O NS^* \<union> NS^*" by auto
    thus "(x, y) \<in>  S O S^* O NS^* \<union> NS^*"
    proof (cases "(x, z) \<in> NS")
      case True
      from compat compat_tr_compat have trCompat: "NS^* O S \<subseteq> S" by blast
      from zy True have "(x, y) \<in> (NS O S O S^* O NS^*) \<union> (NS O NS^*)" by auto
      hence "(x, y) \<in> ((NS O S) O S^* O NS^*) \<union> (NS O NS^*)" by auto
      hence "(x, y) \<in> ((NS^* O S) O S^* O NS^*) \<union> (NS O NS^*)" by auto
      with trCompat have xy:"(x, y) \<in> (S O S^* O NS^*) \<union> (NS O NS^*)" by auto
      have "NS O NS^* \<subseteq> NS^*" by auto
      with xy show "(x, y) \<in> (S O S^* O NS^*) \<union> NS^*" by auto
    next
      case False
      with xz have xz:"(x, z) \<in> S" by auto
      with zy have "(x, y) \<in> S O (S O S^* O NS^* \<union> NS^*)" by auto
      thus "(x, y) \<in> (S O S^* O NS^*) \<union> NS^*" using right_comp_S by simp
    qed
  qed
qed

lemma rtrancl_S_rtrancl_is_NS_comp_S :
  assumes compat: "NS O S \<subseteq> S" 
  shows "(NS \<union> S)^* O S O (NS \<union> S)^* = S O S^* O NS^*"
proof -
  let ?NSuS = "NS \<union> S"
  let ?NSS = "S O S^* O NS^*"
  let ?midS = "?NSuS^* O S O ?NSuS^*"

  have a: "NS^* \<subseteq> ?NSuS^*" using rtrancl_mono by blast
  have "S^* \<subseteq> ?NSuS^*" using rtrancl_mono by blast
  with a have b: "S^* O NS^* \<subseteq> ?NSuS^*" using rtrancl_trans[where r = ?NSuS] by blast
  hence one: "?NSS \<subseteq> ?midS" using rtrancl_refl[where r = ?NSuS] by auto

  have "?NSuS^* O S \<subseteq> (?NSS \<union> NS^*) O S" using split_NS_S_union[where S = S and NS = NS] compat by blast
  also have "\<dots> \<subseteq> ?NSS O S \<union> NS^* O S" by auto
  also have "\<dots> \<subseteq> ?NSS O S \<union> S" using compat compat_tr_compat[where S = S and NS = NS] by auto
  also have "\<dots> \<subseteq> S O ?NSuS^* O S \<union> S" using b by auto
  also have "\<dots> \<subseteq> S O ?NSuS^* O ?NSuS \<union> S" using rtrancl_mono[where r = S and s = ?NSuS] by auto
  also have "\<dots> \<subseteq> S O ?NSuS^* \<union> S" using rtrancl_trans[where r = ?NSuS] rtrancl_refl[where r = ?NSuS] by blast
  also have "\<dots> \<subseteq> S O ?NSuS^*" using rtrancl_refl[where r = ?NSuS] by auto
  finally have "?midS \<subseteq> S O ?NSuS^* O ?NSuS^*" by blast
  also have "\<dots> \<subseteq> S O ?NSuS^*" using rtrancl_trans[where ?r = ?NSuS] by blast
  also have "\<dots> \<subseteq> S O (?NSS \<union> NS^*)" using split_NS_S_union[where S = S and NS = NS] compat by blast
  also have "\<dots> \<subseteq> ?NSS" using right_comp_S[where ?NS = NS and ?S = S] by blast
  finally have two: "?midS \<subseteq> ?NSS" . 

  from one two show ?thesis by auto 
qed

lemma rtrancl_S_rtrancl_is_SN : 
  assumes compat: "NS O S \<subseteq> S" and SN: "SN S"
  shows "SN((NS \<union> S)^* O S O (NS \<union> S)^*)"
using rtrancl_S_rtrancl_is_NS_comp_S[where S = S and NS = NS] 
  NS_comp_S_is_SN[where S = S and NS = NS] assms by force

lemma rtrancl_diff_decomp:
 assumes "(x, y) \<in> A^* - B^*"
 shows "(x, y) \<in> A^* O (A - B) O A^*"
proof-
 from assms have A: "(x, y) \<in> A^*" and B:"(x, y) \<notin> B^*" by auto
 from A have "\<exists> k. (x, y) \<in> A^^k" by (rule rtrancl_imp_rel_pow')
 then obtain k where Ak:"(x, y) \<in> A^^k" by auto
 from Ak B show "(x, y) \<in> A^* O (A - B) O A^*"
 proof (induct k arbitrary: x)
 case 0
  with `(x, y) \<notin> B^*` 0 show ?case using ccontr by auto
 next
 case (Suc i)
  hence B:"(x, y) \<notin> B^*" and ASk:"(x, y) \<in> A ^^ Suc i" by auto
  from ASk have "\<exists>z. (x, z) \<in> A \<and> (z, y) \<in> A ^^ i" using rel_pow_Suc_D2[where ?R=A] by auto
  then obtain z where xz:"(x, z) \<in> A" and "(z, y) \<in> A ^^ i" by auto
  hence zy:"(z, y) \<in> A^*" using rel_pow_imp_rtrancl by auto
  from xz show "(x, y) \<in> A^* O (A - B) O A^*"
  proof (cases "(x, z) \<in> B")
   case False
    with xz zy show "(x, y) \<in> A\<^sup>* O (A - B) O A\<^sup>*" by auto
   next
   case True
    hence "(x, z) \<in> B^*" by auto
    have "\<lbrakk>(x, z) \<in> B^*; (z, y) \<in> B^*\<rbrakk> \<Longrightarrow> (x, y) \<in> B^*" using rtrancl_trans[where ?r=B and ?a=x and ?b=z] by auto
    with  `(x, z) \<in> B^*` `(x, y) \<notin> B^*` have "(z, y) \<notin> B^*" by auto
    with Suc `(z, y) \<in> A ^^ i` have "(z, y) \<in> A^* O (A - B) O A^*" by auto
    with xz have xy:"(x, y) \<in> A O A\<^sup>* O (A - B) O A\<^sup>*" by auto
    have "A O A\<^sup>* \<subseteq> A^*" by auto
    hence "A O A\<^sup>* O (A - B) O A\<^sup>* \<subseteq> A\<^sup>* O (A - B) O A\<^sup>*" by auto
    from this xy show "(x, y) \<in> A\<^sup>* O (A - B) O A\<^sup>*" using subsetD[where ?A="A O A\<^sup>* O (A - B) O A\<^sup>*"] by auto
  qed
 qed
qed

lemma SN_empty[simp]: "SN {}" by auto
lemmas SN_mono = SN_subset

end
