(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Loop imports Trs ErrorMonad begin

text {* abstract formalization of loops *}
types ('f,'v) loop = "('f,'v) term list \<times> ('f,'v) ctxt \<times> ('f,'v) sub"

fun is_rstep_seq:: "('f,'v) term list \<Rightarrow> ('f,'v) trs \<Rightarrow> bool" where
 "is_rstep_seq [t] R = True" |
 "is_rstep_seq (t1#t2#ts) R = ((t1,t2) \<in> rstep R \<and> is_rstep_seq (t2#ts) R)"

fun is_loop:: "('f,'v) loop \<Rightarrow> ('f,'v) trs \<Rightarrow> bool" where
 "is_loop ([],C,\<sigma>) R = False" |
 "is_loop (s#ss,C,\<sigma>) R = is_rstep_seq (s#ss@[C\<langle>s\<cdot>\<sigma>\<rangle>]) R"

fun ith:: "('f,'v) loop \<Rightarrow> nat \<Rightarrow> ('f,'v) term" where
 "ith (t#ts,C,\<sigma>) n = (if n < length (t#ts) then (List.nth (t#ts) n)
                        else C\<langle>(ith (t#ts,C,\<sigma>) (n-length (t#ts)))\<cdot>\<sigma>\<rangle>)"

lemma is_loop_imp_is_rstep_seq [simp]: assumes "is_loop (ts,C,\<sigma>) R" shows "is_rstep_seq (ts@[C\<langle>hd ts\<cdot>\<sigma>\<rangle>]) R" 
 using assms by (cases ts) auto

lemma is_rstep_seq_length: assumes "is_rstep_seq ts R" "Suc i < length ts" shows "(ts ! i, ts ! Suc i) \<in> (rstep R)" 
using assms proof (induct i arbitrary: ts)
 case 0 show ?case 
 proof (cases ts) 
  case Nil thus ?thesis using 0 by auto
 next
  case (Cons s ss) thus ?thesis using 0 by (cases ss) auto
 qed
 next
 case (Suc j) thus ?case
  proof (cases ts) 
  case Nil thus ?thesis using Suc by auto
  next
  case (Cons s ss) hence C: "ts=s#ss" by simp
  show ?thesis
  proof (cases ss) 
   case Nil thus ?thesis using Suc unfolding C by simp
  next
   case (Cons u us) thus ?thesis using Suc unfolding C Cons by force
  qed
 qed
qed

lemma is_loop_length: assumes "is_loop (ts,C,\<sigma>) R" "Suc i < length ts" shows "(ts ! i, ts ! Suc i) \<in> (rstep R)"
proof -
 have A:"Suc i < length (ts@[C\<langle>hd ts\<cdot>\<sigma>\<rangle>])" using assms by simp
 have B:"is_rstep_seq (ts @ [C\<langle>hd ts\<cdot>\<sigma>\<rangle>]) R" using assms by auto 
 have "((ts @ [C\<langle>hd ts\<cdot>\<sigma>\<rangle>]) ! i, (ts @ [C\<langle>hd ts\<cdot>\<sigma>\<rangle>]) ! Suc i) \<in> rstep R" (is "(_,(ts@?YS)!_) \<in> _") using A B is_rstep_seq_length by best
 thus ?thesis using assms List.nth_append[where ys="?YS"] by simp
qed

lemma is_loop_length2: assumes "is_loop (ts,C,\<sigma>) R" "i < length ts" shows "ts ! i = ith (ts,C,\<sigma>) i"
using assms by (cases ts) auto

lemma is_loop_ith [simp]: assumes "is_loop (ts,C,\<sigma>) R" shows "ith (ts,C,\<sigma>) (List.length ts) = C\<langle>hd ts\<cdot>\<sigma>\<rangle>"
using assms by (cases ts) auto

lemma [simp]: assumes "is_loop (ts,C,\<sigma>) R" "Suc n < length ts" shows "(ith (ts,C,\<sigma>) n, ith (ts,C,\<sigma>) (Suc n)) \<in> (rstep R)"
proof -
 have "(ts ! n, ts ! Suc n) \<in> (rstep R)" using assms is_loop_length by best
 thus ?thesis using assms is_loop_length2 by force
qed

lemma is_loop_length_eq [simp]: assumes "is_loop (ts,C,\<sigma>) R" "Suc n = length ts" 
 shows "(ith (ts,C,\<sigma>) n, ith (ts,C,\<sigma>) (Suc n)) \<in> (rstep R)"
proof -
 have A:"Suc n < length (ts@[C\<langle>hd ts\<cdot>\<sigma>\<rangle>])" using assms by auto
 have B: "is_rstep_seq (ts@[C\<langle>(hd ts)\<cdot>\<sigma>\<rangle>]) R" using assms by force
 have C: "n < length ts" using assms by auto
 have "((ts@[C\<langle>(hd ts)\<cdot>\<sigma>\<rangle>]) ! n, (ts@[C\<langle>(hd ts)\<cdot>\<sigma>\<rangle>]) ! (Suc n)) \<in> (rstep R)" using is_rstep_seq_length A B by best
 hence "(ts ! n, (ts@[C\<langle>(hd ts)\<cdot>\<sigma>\<rangle>]) ! (Suc n)) \<in> (rstep R)" using C List.nth_append[where xs="ts"] by auto
 hence "(ith (ts,C,\<sigma>) n, (ts@[C\<langle>(hd ts)\<cdot>\<sigma>\<rangle>]) ! (Suc n)) \<in> (rstep R)" using is_loop_length2 assms C by force
 thus ?thesis using assms by force
qed

lemma is_loop_length_leq: assumes "is_loop (ts,C,\<sigma>) R" "Suc n <= length ts" shows 
 "(ith (ts,C,\<sigma>) n, ith (ts,C,\<sigma>) (Suc n)) \<in> (rstep R)"
proof (cases "Suc n < length ts") 
 case True thus ?thesis using assms by simp
next
 case False thus ?thesis using assms is_loop_length_eq by force
qed

lemma rstep_ctxt_sub_closed [simp]: assumes "(s,t) \<in> (rstep R)" shows "\<forall>C \<sigma>. (C\<langle>s\<cdot>\<sigma>\<rangle>,C\<langle>t\<cdot>\<sigma>\<rangle>) \<in> (rstep R)"
using assms rstep_sub_closed ctxt_closed_rstep_equal_rstep unfolding sub_closed_def by blast

lemma is_loop_length_geq: assumes "is_loop (ts,C,\<sigma>) R" "n >= length ts" shows 
 "ith (ts,C,\<sigma>) n = C\<langle>(ith (ts,C,\<sigma>) (n - length ts))\<cdot>\<sigma>\<rangle>"
using assms by (cases ts) auto

lemma is_loop_length_eq2: assumes "is_loop (ts,C,\<sigma>) R"  shows 
 "ith (ts,C,\<sigma>) (n+length ts) = C\<langle>(ith (ts,C,\<sigma>) n)\<cdot>\<sigma>\<rangle>"
using assms by (cases ts) auto

lemma is_loop_length_gt: assumes "is_loop (ts,C,\<sigma>) R" "Suc n > length ts" shows 
 "(ith (ts,C,\<sigma>) n, ith (ts,C,\<sigma>) (Suc n)) \<in> (rstep R)"
using assms proof (induct n rule:nat_less_induct)
case (1 n)
show ?case
proof (cases "Suc (n-length ts) > length ts")
 case True
  have Z: "length ts > 0" using assms by auto
  have Z1: "n >= length ts"using 1 by arith
  have Z': "Suc n >= length ts"using 1  by arith
  have Z2: "n > 0" using Z Z1 by arith
  have B: "n-length ts < n" using True assms Z by arith
  have "(ith (ts,C,\<sigma>) (n-length ts),ith (ts,C,\<sigma>) (Suc (n-length ts))) \<in> (rstep R)" (is "(?L,?R) \<in> _")
   using 1 True B by fast
  hence "(C\<langle>?L\<cdot>\<sigma>\<rangle>,C\<langle>?R\<cdot>\<sigma>\<rangle>) \<in> (rstep R)" (is "(?L,?R) \<in> _") using rstep_ctxt_sub_closed by best
  hence "(ith (ts,C,\<sigma>) n, ?R) \<in> (rstep R)" using  is_loop_length_geq 1 Z1 by force
  hence "(ith (ts,C,\<sigma>) n,C\<langle>(ith (ts,C,\<sigma>) ((Suc n) - length ts))\<cdot>\<sigma>\<rangle>) \<in> (rstep R)" using Nat.Suc_diff_le Z1 by auto
  thus ?thesis using 1 Z' is_loop_length_geq by force
 next
 case False hence F: "length ts >= Suc (n-length ts)" by simp
  have Z1: "n >= length ts"using 1 by arith
  have "(ith (ts,C,\<sigma>) (n-length ts),ith (ts,C,\<sigma>) (Suc (n-length ts))) \<in> (rstep R)" (is "(?L,?R) \<in> _") 
   using is_loop_length_leq 1 F by best
  hence "(C\<langle>?L\<cdot>\<sigma>\<rangle>,C\<langle>?R\<cdot>\<sigma>\<rangle>)\<in> (rstep R)" (is "(_,?R) \<in> _") using rstep_ctxt_sub_closed 1 by best
  hence "(ith(ts,C,\<sigma>) ((n-length ts)+length ts),?R)\<in> (rstep R)" using is_loop_length_eq2 1(2) by force
  hence "(ith (ts,C,\<sigma>) n,C\<langle>(ith (ts,C,\<sigma>) ((Suc n)-length ts))\<cdot>\<sigma>\<rangle>)\<in> (rstep R)" (is "(?L,_) \<in> _") using Nat.Suc_diff_le Z1 by auto
  hence "(?L,ith (ts,C,\<sigma>) ((Suc n-length ts)+length ts))\<in> (rstep R)" using is_loop_length_geq 1(2) by force
  thus ?thesis using Z1 Nat.Suc_diff_le by simp
 qed
qed

lemma is_loop_imp_ith: assumes "is_loop (ts,C,\<sigma>) R" shows "(ith (ts,C,\<sigma>) n, ith (ts,C,\<sigma>) (Suc n)) \<in> (rstep R)"
proof (cases "Suc n <= length ts") 
 case True thus ?thesis using assms is_loop_length_leq by force
next
 case False thus ?thesis using assms is_loop_length_gt by force
qed

lemma is_loop_implies_not_SN_elt:
  assumes "is_loop (ts,C,\<sigma>) R" shows "\<not> SN_elt (rstep R) (hd ts)" 
proof -
 have A: "\<forall>i. ((ith (ts,C,\<sigma>) i,ith (ts,C,\<sigma>) (Suc i)) \<in> (rstep R))" using assms is_loop_imp_ith by auto
 have B: "ith (ts,C,\<sigma>) 0 = hd ts" using assms by (cases ts) auto
 thus ?thesis using A by auto
qed

theorem is_loop_implies_not_SN: assumes "is_loop (ts,C,\<sigma>) R" shows "\<not> SN (rstep R)"
  using is_loop_implies_not_SN_elt[OF assms] unfolding SN_defs by best


text {*executable code from here to bottom*}

fun rewrites :: "('f::shows,'v::shows)rule \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)rule \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "rewrites (s,t) C \<sigma> (l,r) R = checkB ((s=C\<langle>l\<cdot>\<sigma>\<rangle>) \<and> (t=C\<langle>r\<cdot>\<sigma>\<rangle>) \<and> (l,r) mem R) (''Term '' +#+ (shows s) +@+ '' does not rewrite to term '' +#+ shows t)"

fun rewrites_to ::  
 "(('f::shows,'v::shows)term \<times> ('f,'v)ctxt \<times> ('f,'v)sub \<times> ('f,'v)rule)list \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "rewrites_to [(s,C,\<sigma>,rule)] t R = rewrites (s,t) C \<sigma> rule R"
    | "rewrites_to ((s,C,\<sigma>,rule)#(t,D,\<tau>,rule')#xs) u R = (rewrites (s,t) C \<sigma> rule R \<guillemotright> rewrites_to ((t,D,\<tau>,rule')#xs) u R)"

fun check_loop_d :: "(('f::shows,'v::shows)term \<times> ('f,'v)ctxt \<times> ('f,'v) sub \<times> ('f,'v)rule)list \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "check_loop_d  [] _ _ _ = error(shows_string ''There must be at least one rewrite step'')"
    | "check_loop_d ts D \<tau> R = rewrites_to ts (D\<langle>(fst (hd ts))\<cdot>\<tau>\<rangle>) R"

lemma rewrites_imp_rstep: assumes "isOK(rewrites (s,t) C \<sigma> rule R)" shows "(s,t) \<in> (rstep (set R))" 
 using assms by (cases rule) (auto simp:List.mem_iff) 

declare rewrites.simps[simp del]

lemma rstep_imp_is_rstep_seq [simp]: assumes "(s,t) \<in> (rstep R)" shows "is_rstep_seq [s,t] R" 
 using assms by auto

lemma rewrites_to_imp_is_rstep_seq:
 assumes "isOK(rewrites_to ts t R)" shows "is_rstep_seq (map fst ts@[t]) (set R)"
using assms proof (induct ts)
 case Nil thus ?case by auto
next
 case (Cons s ss) 
 hence IH:"isOK(rewrites_to ss t R) \<Longrightarrow> is_rstep_seq (map fst ss @ [t]) (set R)" "isOK(rewrites_to (s # ss) t R)" by auto
 obtain u C \<sigma> r where decomp:"s=(u,C,\<sigma>,r)" using IH by (cases s) best
 hence d:"isOK(rewrites_to ((u,C,\<sigma>,r)#ss) t R)" using IH by auto
 thus ?case
 proof (cases ss)
  case Nil
  hence "(u,t) \<in> (rstep (set R))" using d unfolding Nil rewrites_to.simps using rewrites_imp_rstep by best
  thus ?thesis unfolding decomp map.simps rstep_imp_is_rstep_seq Nil by force
 next
  case (Cons y ys)
  obtain v D \<tau> r' where decomp2: "y=(v,D,\<tau>,r')" using d unfolding Cons by (cases "y") best
  hence d2: "isOK(rewrites_to (s#y#ys) t R)" using IH Cons by auto
  hence "isOK(rewrites (u,v) C \<sigma> r R) \<and> isOK(rewrites_to (y#ys) t R)" unfolding decomp decomp2 by auto
  hence "(u,v) \<in> (rstep (set R)) \<and> is_rstep_seq (map fst (y#ys)@[t]) (set R)" using rewrites_imp_rstep IH unfolding Cons by best
 thus ?thesis unfolding Cons decomp decomp2 by auto
 qed
qed

declare rewrites_to.simps[simp del]

lemma check_loop_d_imp_is_loop: assumes "isOK(check_loop_d ts D \<tau> R)" shows "is_loop (map fst ts,D,\<tau>) (set R)" 
proof -
 have "isOK(rewrites_to ts (D\<langle>(fst(hd ts))\<cdot>\<tau>\<rangle>) R)" using assms by (cases ts) auto
 hence t:"is_rstep_seq (map fst ts@[(D\<langle>(fst(hd ts))\<cdot>\<tau>\<rangle>)]) (set R)" using rewrites_to_imp_is_rstep_seq by auto
thus ?thesis using assms by (cases ts) auto
qed

theorem check_loop_d_sound: assumes "isOK(check_loop_d ts D \<tau> R)" shows "\<not>SN (rstep (set R))" using assms
proof - 
 have "is_loop (map fst ts,D,\<tau>) (set R)" using assms check_loop_d_imp_is_loop by auto 
 thus ?thesis using is_loop_implies_not_SN by auto
qed

text {* compute details about loop from term list *}

fun decompose_one :: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list option"
  where "decompose_one (Var x) t = Some [(Var x,t)]" |
  "decompose_one s (Var y) = None" |
  "decompose_one (Fun f ts) (Fun g ss) = (if f = g then decomp ts ss else None)"

fun decompose_n :: "nat => ('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list option"
 where 
 "decompose_n 0 s t = decompose_one s t" |
 "decompose_n (Suc n) s t = (case decompose_n n s t of None \<Rightarrow> None | Some ls \<Rightarrow> 
   (foldl (%a (s,t). (case a of None \<Rightarrow> None | Some ys \<Rightarrow>
  (case decompose_one s t of None \<Rightarrow> None | Some xs \<Rightarrow> Some (xs @ ys))))
   (Some [])
   ls))"

fun decompose :: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list option" where
 "decompose s t = decompose_n (Term.sz s) s t"

fun decompose_list :: "(('f,'v) term \<times> ('f,'v) term)list \<Rightarrow> ((('f,'v) term \<times> ('f,'v) term)list) option" where
 "decompose_list ts = (foldl 
  (% a (s,t). (case a of None \<Rightarrow> None | Some ls \<Rightarrow> 
  (case (decompose s t) of None \<Rightarrow> None | Some ys \<Rightarrow> Some (ls @ ys))))
  (Some [])
  ts)"

(* extracting a substitution from a list of pairs *)
fun subst_safe :: "'v list \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list \<Rightarrow> ('f,'v) sub \<Rightarrow> ('f,'v) sub option"
 where "subst_safe vs [] \<sigma> = Some \<sigma>"
    |  "subst_safe vs ((Var x,t) # ts) \<sigma> = 
         (if ((Var x,t) mem ts) then subst_safe vs ts \<sigma> else if x mem vs \<and> ((Var x)\<cdot>\<sigma> ~= t) then None else
          (case subst_safe (x#vs) ts \<sigma> of None \<Rightarrow> None | Some (Sub s) \<Rightarrow>
           Some (Sub (\<lambda> y. if (x = y) then t else s y))))"

fun match_all :: "(('f,'v) term \<times> ('f,'v)term) list \<Rightarrow> ('f,'v) sub option" where
   "match_all ts = 
     (case (decompose_list ts) of None \<Rightarrow> None | Some ls \<Rightarrow> subst_safe [] ls \<iota>)"

fun matching :: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> ('f,'v) sub option" where
 "matching (Var x) t = (if (Var x) = t then Some \<iota> else (Some {x\<leftarrow>t}))" |
 "matching s t = (case (decompose s t) of  None \<Rightarrow> None | Some ls \<Rightarrow> subst_safe [] ls \<iota>)"

fun matches :: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" where
 "matches s t = (case matching s t of None \<Rightarrow> False | Some _ \<Rightarrow> True)"

fun getpr :: "('f::shows,'v::shows)term \<Rightarrow> ('f,'v)term \<Rightarrow> pos \<Rightarrow> ('f,'v)rule \<Rightarrow> 
 (('f,'v)term \<times> ('f,'v)ctxt \<times> ('f,'v)sub \<times> ('f,'v) rule ) option"
where "getpr s t p (l,r) = (case (match_all [(l,s|_ p), (r,t|_p)]) of
  None   \<Rightarrow> None 
| Some \<sigma> \<Rightarrow> 
   (if (isOK(rewrites (s,t) (ctxt_of_pos_term p s) \<sigma> (l,r) [(l,r)])) then 
     Some (s,ctxt_of_pos_term p s,\<sigma>,(l,r)) else None))"

fun getp :: "('f::shows,'v::shows)term \<Rightarrow> ('f,'v)term \<Rightarrow> pos \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
 (('f,'v)term \<times> ('f,'v)ctxt \<times> ('f,'v)sub \<times> ('f,'v) rule ) option"
where "getp s t p [] = None"
    | "getp s t p (r#rs) = (case getpr s t p r of None \<Rightarrow> getp s t p rs | Some d \<Rightarrow> Some d)"

primrec find_get_p ::"('f ::shows,'v ::shows) term \<Rightarrow> ('f,'v) term \<Rightarrow> pos list \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
 (('f,'v)term \<times> ('f,'v)ctxt \<times> ('f,'v)sub \<times> ('f,'v) rule ) option"
where "find_get_p s t [] R = None"
    | "find_get_p s t (p#ps) R = (case (getp s t p R) of None \<Rightarrow> find_get_p s t ps R | Some d \<Rightarrow> Some d)"

fun get :: "('f ::shows,'v ::shows) term \<Rightarrow> ('f,'v) term \<Rightarrow> ('f,'v)trsL \<Rightarrow> (shows,
 (('f,'v)term \<times> ('f,'v)ctxt \<times> ('f,'v)sub \<times> ('f,'v) rule )) emonad"
where "get s t R = (case find_get_p s t ((poss_list s)) R of 
  None     \<Rightarrow>  error (''Could not find reduction between term '' +#+ shows s +@+ '' and term '' +#+ shows t +@+ shows_string ''.'')
| Some res \<Rightarrow> return res
)"

fun get_list :: "('f ::shows,'v ::shows) term list \<Rightarrow> ('f,'v) term \<Rightarrow> ('f,'v) trsL \<Rightarrow> (shows,
 (('f,'v)term \<times> ('f,'v)ctxt \<times> ('f,'v)sub \<times> ('f,'v) rule) list) emonad"
where "get_list [] u R = error(shows_string ''Loop must be non-empty'')" 
    | "get_list [s] u R = (get s u R \<rightarrow> (\<lambda> d. [d]))" 
    | "get_list (s#t#ts) u R = (get s t R \<guillemotright>= (\<lambda> d. get_list (t#ts) u R \<rightarrow> (\<lambda> ls. d # ls)))"

fun get_loop :: "('f ::shows,'v ::shows) term list \<Rightarrow> ('f,'v) ctxt \<Rightarrow> ('f,'v) sub \<Rightarrow> ('f,'v) trsL \<Rightarrow> (shows,
 ((('f,'v)term \<times> ('f,'v)ctxt \<times> ('f,'v)sub \<times> ('f,'v) rule) list)) emonad"
where "get_loop [] _ _ _ = error(shows_string ''Loop must be non-empty'')"
    | "get_loop (t#ts) D \<tau> R = get_list (t#ts) D\<langle>t\<cdot>\<tau>\<rangle> R"

fun check_loop :: "(('f ::shows ,'v ::shows) term list \<times> ('f,'v) ctxt \<times> ('f,'v) sub) \<Rightarrow> ('f,'v) trsL \<Rightarrow> shows monad"
where "check_loop (ts,C,\<sigma>) R = (get_loop ts C \<sigma> R \<guillemotright>= (\<lambda> ls. check_loop_d ls C \<sigma> R))"

text {* boring properties about the functions *}
lemma inc: assumes "isOK(rewrites (s,t) C \<sigma> rule R)" shows "isOK(rewrites (s,t) C \<sigma> rule (S@R@T))"
proof -
 from assms obtain l r where x: "rule=(l,r)" by (cases rule) auto 
 hence "isOK(rewrites (s,t) C \<sigma> (l,r) R)" using assms by auto
 hence "isOK(rewrites (s,t) C \<sigma> (l,r) (S@R@T))" unfolding rewrites.simps mem_iff by auto 
 thus ?thesis using x by auto
qed

lemma getpr_decompose_rule: 
  assumes "getpr s t p rule = Some (x,D,\<tau>,rule')" shows "rule=rule'"
proof -
  obtain l r where d: "rule = (l,r)" by (cases rule) auto thus ?thesis
  proof (cases "match_all [(l,s|_p),(r,t|_p)]")
    case None thus ?thesis using assms d by auto
  next
    case (Some a)
    have cond: "(if isOK(rewrites (s,t) (ctxt_of_pos_term p s) a (l,r) [(l,r)]) then Some (s,ctxt_of_pos_term p s,a,l,r) else None) =
    Some (x, D, \<tau>, rule')" (is "(if ?C then _ else _)= _ ") using assms unfolding d getpr.simps Some by auto
    thus ?thesis unfolding rewrites.simps using d by (cases ?C) (auto simp: rewrites.simps)
  qed
qed

lemma getpr_decompose_term: assumes "getpr s t p rule = Some (x,C,\<sigma>,rule)" shows "(x=s)"
proof -
 obtain l r where d:"rule=(l,r)" using assms by (cases rule) auto
 thus ?thesis 
  (*proof (cases "matching l s|_p")*)
  proof (cases "match_all [(l,s|_p),(r,t|_p)]")
  case None thus ?thesis using assms unfolding d getpr.simps by auto
 next
  case (Some a) thus ?thesis using assms unfolding d by (cases "isOK(rewrites (s,t) (ctxt_of_pos_term p s) a (l,r) [(l,r)])") auto 
 qed
qed

lemma getp_decompose: assumes "getp s t p R = Some (x,C,\<sigma>,rule)" shows "(x=s)" using assms
proof (induct R)
 case Nil thus ?case by auto
next
 case (Cons rule' rs)
 obtain l r where d: "rule'=(l,r)" by (cases rule) force 
 thus ?thesis
 proof (cases "getpr s t p rule'")
  case None thus ?thesis using Cons by auto
 next
  case (Some a)
  have "getpr s t p rule' = Some (x,C,\<sigma>,rule)" using Cons Some by auto
  thus ?thesis using getpr_decompose_term getpr_decompose_rule by fast
 qed
qed

lemma find_get_p_decompose: assumes "find_get_p s t ps R = Some (x,C,\<sigma>,r)" shows "x=s"
using assms proof (induct ps)
 case Nil thus ?case by auto
next
 case (Cons p ps') thus ?case using getp_decompose by (cases "getp s t p R") auto
qed


lemma get_decompose: assumes "get s t R = return (x,C,\<sigma>,r)" shows "x=s" using assms find_get_p_decompose unfolding get.simps
 by (cases "find_get_p s t ((poss_list s)) R", auto simp:return_def error_def, blast) 
(* TODO: adding return_def and error_def should not be required, but corresponding lemmas should 
   be added in ErrorMonad.thy *)

lemma getpr_imp_rewrites: assumes "getpr s t p rule = Some (x,D,\<tau>,rule)" shows "isOK(rewrites (x,t) D \<tau> rule ([rule]))" 
proof -
 obtain l r where d:"rule=(l,r)" by (cases "rule") auto
 show ?thesis
(* proof (cases "matching l s |_p")*)
 proof (cases "match_all [(l,s|_p),(r,t|_p)]")
  case None thus ?thesis using assms unfolding d by simp
 next
  case (Some a) thus ?thesis using assms unfolding d by (cases "isOK(rewrites (s,t) (ctxt_of_pos_term p s) a (l,r) [(l,r)])") auto
 qed
qed

lemma getp_imp_rewrites: assumes "getp s t p R = Some (x,C,\<sigma>,rule)" shows "isOK(rewrites (x,t) C \<sigma> rule R)"
using assms
proof (induct R)
 case Nil thus ?case by auto
next
 case (Cons r RS) thus ?case
 proof (cases "getpr s t p r") 
  case None thus ?thesis using Cons None inc[where S="[r]" and T="[]"] by auto
 next
  case (Some a)
  hence "Some a = Some (x, C, \<sigma>, rule)" using Cons Some by auto
  hence eq: "r = rule" using Some getpr_decompose_rule by best
  from Some have y: "getpr s t p rule = Some (x, C, \<sigma>, rule)" using Cons eq by auto
  hence "isOK(rewrites (x,t) C \<sigma> rule [rule])" using getpr_imp_rewrites by best
  hence "isOK(rewrites (x,t) C \<sigma> rule ([]@[rule]@RS))" using inc eq by best
  thus ?thesis unfolding eq by auto
 qed
qed

lemma find_get_p_imp_rewrites: assumes "find_get_p s t ps R = Some (x,C,\<sigma>,r)" shows "isOK(rewrites (x,t) C \<sigma> r R)"
using assms
proof (induct ps)
 case Nil thus ?case by auto
next
 case (Cons p ps') show ?case 
 proof (cases "getp s t p R")
  case None thus ?thesis using Cons by auto
 next
  case (Some a) 
  have "getp s t p R = Some (x, C, \<sigma>, r)" using Cons Some by auto
  thus ?thesis using getp_imp_rewrites by best
 qed
qed

lemma get_imp_rewrites: assumes "get s t R = return (x,C,\<sigma>,rule)" shows "isOK(rewrites (x,t) C \<sigma> rule R)"
 using assms find_get_p_imp_rewrites unfolding get.simps 
 by (cases "find_get_p s t ((poss_list s)) R", auto simp:return_def error_def, blast) 
(* TODO: adding return_def and error_def should not be required, but corresponding lemmas should 
   be added in ErrorMonad.thy *)

lemma get_imp_rstep: assumes "get s t R = return a" shows "(s,t) \<in> (rstep (set R))" 
 using assms get_decompose get_imp_rewrites rewrites_imp_rstep by (cases a) fast

declare get.simps[simp del]

lemma get_imp_rstep2: assumes "isOK(get s t R)" shows "(s,t) \<in> (rstep (set R))"
using get_imp_rstep assms unfolding return_def 
 by (cases "get s t R", auto simp:isValue_def, blast)
(* TODO: adding return_def and isValue_def should not be required, but corresponding lemmas should 
   be added in ErrorMonad.thy *)


lemma get_list_imp_is_rstep_seq[simp]: assumes "isOK(get_list (t#ts) u R)" shows "is_rstep_seq (t#ts@[u]) (set R)"
using assms proof (induct ts arbitrary:t)
 case Nil thus ?case by (simp add: get_imp_rstep2)
next 
 case (Cons s ss) 
 hence d1: "isOK(get t s R)" by (cases ss, auto simp:bindSound)
 have d2: "isOK(get_list (s#ss) u R)" using Cons by (simp add:bindSound, cases "get t s R", auto simp: isValue_def return_def)
 hence "(t,s) \<in> (rstep (set R)) \<and> is_rstep_seq (s#ss@[u]) (set R)" using d1 get_imp_rstep2 Cons bindSound by fast
 thus ?case by auto
qed
(* TODO: adding return_def and isValue_def should not be required, but corresponding lemmas should 
   be added in ErrorMonad.thy *)

declare get_list.simps[simp del]

lemma get_loop_imp_is_loop[simp]: assumes "isOK(get_loop ts D \<tau> R)" shows "is_loop (ts,D,\<tau>) (set R) "
proof (cases ts)
 case Nil thus ?thesis using assms by auto
next
 case (Cons s ss)
  hence "is_rstep_seq (ts@[D\<langle>s\<cdot>\<tau>\<rangle>]) (set R)" using assms by simp 
  hence "is_loop (ts,D,\<tau>) (set R)" unfolding Cons using is_loop.simps by auto
 thus ?thesis using is_loop_implies_not_SN by auto 
qed

declare get_loop.simps[simp del]

lemma check_loop_imp_is_loop: assumes "isOK(check_loop (ts,D,\<tau>) R)" shows "is_loop (ts,D,\<tau>) (set R)" 
 using assms by (simp add:bindSound) 

theorem check_loop_sound[simp]: assumes "isOK(check_loop (ts,D,\<tau>) R)" shows "\<not> SN (rstep (set R))"
proof -
 have "is_loop (ts,D,\<tau>) (set R)" using assms by (simp add:bindSound)
 thus ?thesis using is_loop_implies_not_SN by auto
qed

declare check_loop.simps[simp del]

text {* appropriate wrappers for loop *}
types ('f,'v) subL = "(('f,'v)term \<times>('f,'v)term) list"
types ('f,'v) loopL = "('f,'v)term list \<times> ('f,'v)ctxt \<times> ('f,'v)subL"

fun "checkLoop" :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)loopL \<Rightarrow> shows monad"
where "checkLoop R (ts,C,sub) = ((case Loop.subst_safe [] sub \<iota> of
  None   \<Rightarrow> error(shows_string ''substitution not unique'')
| Some \<sigma> \<Rightarrow> check_loop (ts,C,\<sigma>) R
) \<leftarrow> (\<lambda>s. ''error in certification of loop'' +#+ shows_nl +@+ s)
)"

end
